"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.
"""

import csv
import datetime as dt
import hashlib
import json
import logging
import math
import os
import pathlib as pl
import re
import socket
import subprocess
import sys
import urllib.parse
import wave
from decimal import Decimal as dec
from decimal import getcontext, ROUND_DOWN
from shutil import copyfile
from typing import Union, Tuple

import numpy as np
from PyQt5.QtGui import QPixmap, QImage

from PIL.ImageQt import Image

from . import config as cfg

try:
    from . import mpv2 as mpv

    # check if MPV API v. 1
    # is v. 1 use the old version of mpv.py
    try:
        if "libmpv.so.1" in mpv.sofile:
            from . import mpv as mpv
    except AttributeError:
        if "mpv-1.dll" in mpv.dll:
            from . import mpv as mpv

except RuntimeError:  # libmpv found but version too old
    from . import mpv as mpv


def mpv_lib_version() -> Tuple[str, str]:
    """
    Version of MPV library

    Returns:
        str: MPV library version
    """
    mpv_lib_file = None
    if sys.platform.startswith("linux"):
        mpv_lib_file = mpv.sofile
    if sys.platform.startswith("win"):
        mpv_lib_file = mpv.dll

    return (".".join([str(x) for x in mpv._mpv_client_api_version()]), mpv_lib_file)


def python_mpv_script_version() -> str:
    """
    version of python-mpv script
    """
    try:
        return mpv.__version__
    except Exception:
        return "Not found"


def error_info(exc_info: tuple) -> tuple:
    """
    return details about error
    usage: error_info(sys.exc_info())

    Args:
        sys.exc_info() (tuple):

    Returns:
        tuple: error type, error file name, error line number
    """

    exc_type, exc_obj, exc_tb = exc_info
    fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]

    return (f"{exc_type}: {exc_obj}", fname, exc_tb.tb_lineno)


def pil2pixmap(im: Image) -> QPixmap:
    """
    convert PIL image to pixmap
    see https://stackoverflow.com/questions/34697559/pil-image-to-qpixmap-conversion-issue
    """

    if im.mode == "RGB":
        r, g, b = im.split()
        im = Image.merge("RGB", (b, g, r))
    elif im.mode == "RGBA":
        r, g, b, a = im.split()
        im = Image.merge("RGBA", (b, g, r, a))
    elif im.mode == "L":
        im = im.convert("RGBA")

    im2 = im.convert("RGBA")
    data = im2.tobytes("raw", "RGBA")
    qim = QImage(data, im.size[0], im.size[1], QImage.Format_ARGB32)
    pixmap = QPixmap.fromImage(qim)
    return pixmap


def replace_leading_trailing_chars(s: str, old_char: str, new_char: str) -> str:
    """
    replace leading and trailing old_char by new_char

    Args:
        s: string
        old_char: character to be replaced
        new_char: character for replacing

    Returns:
        str: string with characters replaced
    """

    sp = s.partition(s.strip(old_char))

    return f"{sp[0].replace(old_char, new_char)}{sp[1]}{sp[2].replace(old_char, new_char)}"


def return_file_header(file_name: str, row_number: int = 5) -> list:
    """
    return file header

    Args:
        file_name (str): path of file
        row_number (int): number of rows to return

    Returns:
        list: first row_number row(s) of file_name
    """
    header: list = []
    try:
        with open(file_name) as f_in:
            for _ in range(row_number):
                header.append(f_in.readline())
    except Exception:
        return []
    return header


def return_file_header_footer(file_name: str, file_row_number: int = 0, row_number: int = 5) -> Tuple[list, list]:
    """
    return file header and footer

    Args:
        file_name (str): path of file
        file_row_number (int): total rows number of file
        row_number (int): number of rows to return

    Returns:
        list: first row_number row(s) of file_name
    """
    header: list = []
    footer: list = []
    try:
        row_idx: int = 0
        with open(file_name, "rt") as f_in:
            for row in f_in:
                if row_idx < row_number:
                    header.append(row.strip())
                if file_row_number > row_number * 2 and (row_idx >= file_row_number - row_number):
                    footer.append(row.strip())
                row_idx += 1

    except Exception:
        return [], []
    return header, footer


def bytes_to_str(b: bytes) -> str:
    """
    Translate bytes to string.

    Args:
        b (bytes): byte to convert

    Returns:
        str: converted byte
    """

    if isinstance(b, bytes):
        fileSystemEncoding = sys.getfilesystemencoding()
        # hack for PyInstaller
        if fileSystemEncoding is None:
            fileSystemEncoding = "UTF-8"
        return b.decode(fileSystemEncoding)
    else:
        return b


def convertTime(time_format: str, sec: Union[float, dec]) -> Union[str, None]:
    """
    convert time in base at the current format (S or HHMMSS)

    Args:
        sec (float): time in seconds

    Returns:
        string: time in base of current format (self.timeFormat S or cfg.HHMMSS)
    """

    if isinstance(sec, dec) and sec.is_nan():
        return cfg.NA

    if time_format == cfg.S:
        return f"{sec:.3f}"

    if time_format == cfg.HHMMSS:
        return seconds2time(sec)

    return None


def smart_time_format(sec: Union[float, dec], time_format: str = cfg.S, cutoff: dec = cfg.SMART_TIME_CUTOFF_DEFAULT) -> str:
    """
    Smart time format
    returns time in seconds if <= cutoff else in HH:MM:SS.ZZZ format
    """
    # cutoff = 0 follows the time format selectd by user
    if cutoff == 0:
        return convertTime(time_format, sec)
    if sec <= cutoff:
        return f"{sec:.3f}"
    else:
        return seconds2time(sec)


def convert_time_to_decimal(pj: dict) -> dict:
    """
    convert time of project from float to decimal

    Args:
        pj (dict): BORIS project

    Returns:
        dict: BORIS project
    """
    for obs_id in pj[cfg.OBSERVATIONS]:
        if cfg.TIME_OFFSET in pj[cfg.OBSERVATIONS][obs_id]:
            if pj[cfg.OBSERVATIONS][obs_id][cfg.TIME_OFFSET] is not None:
                pj[cfg.OBSERVATIONS][obs_id][cfg.TIME_OFFSET] = dec(str(pj[cfg.OBSERVATIONS][obs_id][cfg.TIME_OFFSET]))
            else:
                pj[cfg.OBSERVATIONS][obs_id][cfg.TIME_OFFSET] = dec("0.000")
        for idx, _ in enumerate(pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS]):
            pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS][idx][cfg.EVENT_TIME_FIELD_IDX] = dec(
                pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS][idx][cfg.EVENT_TIME_FIELD_IDX]
            ).quantize(dec(".001"))

    return pj


def count_media_file(media_files: dict) -> int:
    """
    count number of media file for observation
    """
    return sum([len(media_files[idx]) for idx in media_files])


def file_content_md5(file_name: str) -> str:
    """
    returns the MD5 sum of file content
    """
    hash_md5 = hashlib.md5()
    try:
        with open(file_name, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()
    except FileNotFoundError:
        return ""


def txt2np_array(
    file_name: str, columns_str: str, substract_first_value: str, converters=None, column_converter=None
) -> Tuple[bool, str, np.array]:
    """
    read a txt file (tsv or csv) and return a np array with columns cited in columns_str

    Args:
        file_name (str): path of the file to load in numpy array
        columns_str (str): indexes of columns to be loaded. First columns must be the timestamp. Example: "4,5"
        substract_first_value (str): "True" or "False"
        converters (dict): dictionary containing converters
        column_converter (dict): dictionary key: column index, value: converter name

    Returns:
        bool: True if data successfullly loaded, False if case of error
        str: error message. Empty if success
        numpy array: data. Empty if not data failed to be loaded

    """
    if converters is None:
        converters = {}
    if column_converter is None:
        column_converter = {}

    # check columns
    try:
        columns = [int(x) - 1 for x in columns_str.split(",")]
    except Exception:
        return False, f"Problem with columns {columns_str}", np.array([])

    # check converters
    np_converters: dict = {}
    for column_idx in column_converter:
        if column_converter[column_idx] in converters:
            conv_name = column_converter[column_idx]

            function = f"""def {conv_name}(INPUT):\n"""
            function += """    INPUT = INPUT.decode("utf-8") if isinstance(INPUT, bytes) else INPUT"""
            for line in converters[conv_name]["code"].split("\n"):
                function += f"    {line}\n"
            function += """    return OUTPUT"""

            try:
                exec(function)
            except Exception:
                return False, f"error in converter: {sys.exc_info()[1]}", np.array([])

            np_converters[column_idx - 1] = locals()[conv_name]

        else:
            return False, f"converter {cfg.converters_param[column_idx]} not found", np.array([])

    # snif txt file
    try:
        with open(file_name) as csvfile:
            buff = csvfile.read(4096)
            snif = csv.Sniffer()
            dialect = snif.sniff(buff)
            """has_header = snif.has_header(buff)"""
        # count number of header rows
        header_rows_nb = 0
        csv.register_dialect("dialect", dialect)
        with open(file_name, "r") as f:
            reader = csv.reader(f, dialect="dialect")
            for row in reader:
                if sum([isinstance(intfloatstr(x), str) for x in row]) == len(row):
                    header_rows_nb += 1

    except Exception:
        return False, f"{sys.exc_info()[1]}", np.array([])

    try:
        data = np.loadtxt(file_name, delimiter=dialect.delimiter, usecols=columns, skiprows=header_rows_nb, converters=np_converters)

    except Exception:
        return False, f"{sys.exc_info()[1]}", np.array([])

    # check if first value must be substracted
    if substract_first_value == "True":
        data[:, 0] -= data[:, 0][0]

    return True, "", data


def versiontuple(version_str: str) -> tuple:
    """
    Convert version from str to tuple of str

    Args:
        version_str (str): version

    Returns:
        tuple[str, str, str]: version in tuple format (for comparison)
    """
    filled = []
    for point in version_str.split("."):
        filled.append(point.zfill(8))
    return tuple(filled)


def behavior_user_color(ethogram: dict, behavior_code: str) -> Union[str, None]:
    """
    returns the color of behavior if defined else None
    """
    for x in ethogram:
        if ethogram[x][cfg.BEHAVIOR_CODE] == behavior_code:
            if ethogram[x].get(cfg.COLOR, None) == "":
                return None
            else:
                return ethogram[x].get(cfg.COLOR, None)

    return None


def behav_category_user_color(behavioral_categories: dict, name: str) -> Union[str, None]:
    """
    returns the color of the behavioral category if defined else None
    """
    for key in behavioral_categories:
        if behavioral_categories[key]["name"] == name:
            return behavioral_categories[key].get(cfg.COLOR, None)

    return None


def state_behavior_codes(ethogram: dict) -> list:
    """
    behavior codes defined as STATE event

    Args:
        ethogram (dict): ethogram dictionary

    Returns:
        list: list of behavior codes defined as STATE event

    """
    return [ethogram[x][cfg.BEHAVIOR_CODE] for x in ethogram if cfg.STATE in ethogram[x][cfg.TYPE].upper()]


def point_behavior_codes(ethogram: dict) -> list:
    """
    behavior codes defined as POINT event

    Args:
        ethogram (dict): ethogram dictionary

    Returns:
        list: list of behavior codes defined as POINT event

    """
    return [ethogram[x][cfg.BEHAVIOR_CODE] for x in ethogram if cfg.POINT in ethogram[x][cfg.TYPE].upper()]


def group_events(pj: dict, obs_id: str, include_modifiers: bool = False) -> dict:
    """
    group events by subject, behavior, modifier (if required)

    result is a dict like:

    {(subject, behavior, ""): list of tuple (start: Decimal, end: Decimal)}

    or with modifiers:

    {(subject, behavior, modifier): list of tuple (start: Decimal, end: Decimal)}

    in case of point events start=end
    """

    try:
        state_events_list = state_behavior_codes(pj[cfg.ETHOGRAM])
        point_events_list = point_behavior_codes(pj[cfg.ETHOGRAM])
        mem_behav = {}
        intervals_behav = {}

        for event in pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS]:
            time_ = event[cfg.EVENT_TIME_FIELD_IDX]
            subject = event[cfg.EVENT_SUBJECT_FIELD_IDX]
            code = event[cfg.EVENT_BEHAVIOR_FIELD_IDX]
            modifier = event[cfg.EVENT_MODIFIER_FIELD_IDX] if include_modifiers else ""

            # check if code is state
            if code in state_events_list:
                if (subject, code, modifier) in mem_behav and mem_behav[(subject, code, modifier)]:
                    if (subject, code, modifier) not in intervals_behav:
                        intervals_behav[(subject, code, modifier)] = []
                    intervals_behav[(subject, code, modifier)].append((mem_behav[(subject, code, modifier)], time_))

                    mem_behav[(subject, code, modifier)] = 0
                else:
                    mem_behav[(subject, code, modifier)] = time_

            # check if code is state
            if code in point_events_list:
                if (subject, code, modifier) not in intervals_behav:
                    intervals_behav[(subject, code, modifier)] = []
                intervals_behav[(subject, code, modifier)].append((time_, time_))

        return intervals_behav

    except Exception:
        return {"error": ""}


def get_current_states_modifiers_by_subject(
    state_behaviors_codes: list, events: list, subjects: dict, time_: dec, include_modifiers: bool = False
) -> dict:
    """
    get current states and modifiers (if requested) for subjects at given time

    Args:
        state_behaviors_codes (list): list of behavior codes defined as STATE event
        events (list): list of events
        subjects (dict): dictionary of subjects
        time (Decimal): time or image index for an observation from images
        include_modifiers (bool): include modifier if True (default: False)

    Returns:
        dict: current states by subject. dict of list
    """
    current_states: dict = {}
    if time_.is_nan():
        for idx in subjects:
            current_states[idx] = []
        return current_states

    # check if time contains NA
    if [x for x in events if x[cfg.EVENT_TIME_FIELD_IDX].is_nan()]:
        check_index = cfg.PJ_OBS_FIELDS[cfg.IMAGES][cfg.IMAGE_INDEX]
    else:
        check_index = cfg.EVENT_TIME_FIELD_IDX

    if include_modifiers:
        for idx in subjects:
            current_states[subjects[idx]["name"]] = {}
        for x in events:
            if x[check_index] > time_:
                break
            if x[cfg.EVENT_BEHAVIOR_FIELD_IDX] in state_behaviors_codes:
                if (x[cfg.EVENT_BEHAVIOR_FIELD_IDX], x[cfg.EVENT_MODIFIER_FIELD_IDX]) not in current_states[x[cfg.EVENT_SUBJECT_FIELD_IDX]]:
                    current_states[x[cfg.EVENT_SUBJECT_FIELD_IDX]][
                        (x[cfg.EVENT_BEHAVIOR_FIELD_IDX], x[cfg.EVENT_MODIFIER_FIELD_IDX])
                    ] = False

                current_states[x[cfg.EVENT_SUBJECT_FIELD_IDX]][
                    (x[cfg.EVENT_BEHAVIOR_FIELD_IDX], x[cfg.EVENT_MODIFIER_FIELD_IDX])
                ] = not current_states[x[cfg.EVENT_SUBJECT_FIELD_IDX]][(x[cfg.EVENT_BEHAVIOR_FIELD_IDX], x[cfg.EVENT_MODIFIER_FIELD_IDX])]

        r: dict = {}
        for idx in subjects:
            r[idx] = [f"{bm[0]} ({bm[1]})" for bm in current_states[subjects[idx]["name"]] if current_states[subjects[idx]["name"]][bm]]

    else:
        for idx in subjects:
            current_states[subjects[idx]["name"]] = {}
            for b in state_behaviors_codes:
                current_states[subjects[idx]["name"]][b] = False
        for x in events:
            if x[check_index] > time_:
                break
            if x[cfg.EVENT_BEHAVIOR_FIELD_IDX] in state_behaviors_codes:
                current_states[x[cfg.EVENT_SUBJECT_FIELD_IDX]][x[cfg.EVENT_BEHAVIOR_FIELD_IDX]] = not current_states[
                    x[cfg.EVENT_SUBJECT_FIELD_IDX]
                ][x[cfg.EVENT_BEHAVIOR_FIELD_IDX]]

        r: dict = {}
        for idx in subjects:
            r[idx] = [b for b in state_behaviors_codes if current_states[subjects[idx]["name"]][b]]

    return r


def get_current_states_modifiers_by_subject_2(state_behaviors_codes: list, events: list, subjects: dict, time: dec) -> dict:
    """
    get current states and modifiers for subjects at given time
    differs from get_current_states_modifiers_by_subject in the output format: [behavior, modifiers]

    Args:
        state_behaviors_codes (list): list of behavior codes defined as STATE event
        events (list): list of events
        subjects (dict): dictionary of subjects
        time (Decimal): time

    Returns:
        dict: current states by subject. dict of list
    """
    current_states = {}
    for idx in subjects:
        current_states[idx] = []
        for sbc in state_behaviors_codes:
            bl = [
                (x[cfg.EVENT_BEHAVIOR_FIELD_IDX], x[cfg.EVENT_MODIFIER_FIELD_IDX])
                for x in events
                if x[cfg.EVENT_SUBJECT_FIELD_IDX] == subjects[idx][cfg.SUBJECT_NAME]
                and x[cfg.EVENT_BEHAVIOR_FIELD_IDX] == sbc
                and x[cfg.EVENT_TIME_FIELD_IDX] <= time
            ]

            if len(bl) % 2:  # test if odd
                current_states[idx].append(bl[-1])

    return current_states


def get_current_points_by_subject(
    point_behaviors_codes: list,
    events: list,
    subjects: dict,
    time: dec,
    tolerance: dec,
    include_modifiers: bool = False,
) -> dict:
    """
    get point events for subjects between given time (time) and (time + tolerance)
    includes modifiers
    Args:
        point_behaviors_codes (list): list of behavior codes defined as POINT event
        events (list): list of events
        subjects (dict): dictionary of subjects
        time (Decimal): time (s)
        tolerance (Decimal): tolerance (s)
        include_modifiers (bool): True to include modifiers

    Returns:
        dict: current point behaviors by subject. dict of list
    """

    current_points = {}
    for idx in subjects:
        current_points[idx] = []
        for sbc in point_behaviors_codes:
            # if include_modifiers:
            point_events = [
                (x[cfg.EVENT_BEHAVIOR_FIELD_IDX], x[cfg.EVENT_MODIFIER_FIELD_IDX])
                for x in events
                if x[cfg.EVENT_SUBJECT_FIELD_IDX] == subjects[idx]["name"]
                and x[cfg.EVENT_BEHAVIOR_FIELD_IDX] == sbc
                # and abs(x[EVENT_TIME_FIELD_IDX] - time) <= tolerance
                and time <= x[cfg.EVENT_TIME_FIELD_IDX] < (time + tolerance)
            ]

            # else:
            #    point_events = [x[EVENT_BEHAVIOR_FIELD_IDX] for x in events
            #                    if x[EVENT_SUBJECT_FIELD_IDX] == subjects[idx]["name"]
            #                    and x[EVENT_BEHAVIOR_FIELD_IDX] == sbc
            #                   # and abs(x[EVENT_TIME_FIELD_IDX] - time) <= tolerance
            #                    and time <= x[EVENT_TIME_FIELD_IDX] < (time + tolerance)]
            for point_event in point_events:
                current_points[idx].append(point_event)

    return current_points


def get_ip_address():
    """Get current IP address

    Args:

    Returns:
        str: IP address
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    return s.getsockname()[0]


def check_txt_file(file_name: str) -> dict:
    """
    Extract parameters of txt file (test for tsv csv)

    Args:
        filename (str): path of file to be analyzed

    Returns:
        dict: {"homogeneous": True or False,
               "fields_number": number of fields,
               "separator": separator char
              }
    """
    try:
        # snif txt file
        with open(file_name) as csvfile:
            buff = csvfile.read(4096)
            snif = csv.Sniffer()
            dialect = snif.sniff(buff)
            has_header = snif.has_header(buff)

        csv.register_dialect("dialect", dialect)
        rows_len: list = []
        with open(file_name, "r") as f:
            reader = csv.reader(f, dialect="dialect")
            for row in reader:
                if not row:
                    continue
                """
                if len(row) not in rows_len:
                    rows_len.append(len(row))
                    if len(rows_len) > 1:
                        break
                """
                rows_len.append(len(row))

        rows_number = len(rows_len)
        rows_uniq_len = set(rows_len)

        # test if file empty
        if not rows_uniq_len:
            return {"error": "The file is empty"}

        if len(rows_uniq_len) == 1:
            return {
                "homogeneous": True,
                "fields number": rows_len[0],
                "separator": dialect.delimiter,
                "rows number": rows_number,
                "has header": has_header,
            }
        else:
            return {"homogeneous": False}
    except Exception:
        return {"error": str(sys.exc_info()[1])}


def extract_wav(ffmpeg_bin: str, media_file_path: str, tmp_dir: str) -> str:
    """
    extract wav from media file and save file in tmp_dir

    Args:
        media_file_path (str): media file path
        tmp_dir (str): temporary dir where to save wav file

    Returns:
        str: wav file path or "" if error
    """

    wav_file_path = pl.Path(tmp_dir) / pl.Path(media_file_path + ".wav").name

    # check if media file is a wav file
    try:
        wav = wave.open(media_file_path, "r")
        wav.close()
        logging.debug(f"{media_file_path} is a WAV file. Copying in the temp directory...")
        copyfile(media_file_path, wav_file_path)
        logging.debug(f"{media_file_path} copied in {wav_file_path}")
        return str(wav_file_path)
    except Exception:
        if wav_file_path.is_file():
            return str(wav_file_path)
        # extract wav file using FFmpeg

        p = subprocess.Popen(
            f'"{ffmpeg_bin}" -i "{media_file_path}" -y -ac 1 -vn "{wav_file_path}"',
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
        )
        out, error = p.communicate()
        out, error = out.decode("utf-8"), error.decode("utf-8")
        logging.debug(f"{out}, {error}")

        if "does not contain any stream" not in error:
            if wav_file_path.is_file():
                return str(wav_file_path)
            return ""
        else:
            return ""


def decimal_default(obj):
    if isinstance(obj, dec):
        return float(round(obj, 3))
    raise TypeError


def complete(lst: list, max_: int) -> list:
    """
    complete list with empty string ("") until len = max

    Args:
        lst (list): list to complete
        max_ (int): number of items to reach

    Returns:
        list: list completed to max_ items with empty string ("")
    """
    while len(lst) < max_:
        lst.append("")
    return lst


def datetime_iso8601(dt) -> str:
    """
    current date time in ISO8601 format without microseconds
    example: 2019-06-13 10:01:02

    Returns:
        str: date time in ISO8601 format without microseconds
    """
    return dt.isoformat(sep=" ", timespec="seconds")


def seconds_of_day(timestamp: dt.datetime) -> dec:
    """
    return the number of seconds since start of the day

    Returns:
        dev: number of seconds since the start of the day
    """

    # logging.debug("function: seconds_of_day")
    # logging.debug(f"{timestamp = }")

    t = timestamp.time()
    return dec(t.hour * 3600 + t.minute * 60 + t.second + t.microsecond / 1000000).quantize(dec("0.001"))


def sorted_keys(d: dict) -> list:
    """
    return list of sorted keys of provided dictionary

    Args:
        d (dict): dictionary

    Returns:
         list: dictionary keys sorted numerically
    """
    return [str(x) for x in sorted([int(x) for x in d.keys()])]


def intfloatstr(s: str) -> int:
    """
    convert str in int or float or return str
    """

    try:
        return int(s)
    except Exception:
        try:
            return f"{float(s):0.3f}"
        except Exception:
            return s


def distance(p1, p2):
    """
    euclidean distance between 2 points
    """
    x1, y1 = p1
    x2, y2 = p2
    return ((x1 - x2) ** 2 + (y1 - y2) ** 2) ** 0.5


def angle(vertex: tuple, side1: tuple, side2: tuple) -> float:
    """
    Determine the angle between 3 points (p1 must be the vertex)
    return angle in degree

    Args:
        vertex (tuple): vertex
        side1 (tuple): side 1
        side2 (tuple): side 2

    Returns:
        float: angle between side1 - vertex - side2
    """
    return (
        math.acos(
            (distance(vertex, side1) ** 2 + distance(vertex, side2) ** 2 - distance(side1, side2) ** 2)
            / (2 * distance(vertex, side1) * distance(vertex, side2))
        )
        / math.pi
        * 180
    )


def oriented_angle(P1: tuple, P2: tuple, P3: tuple) -> float:
    """
    Calculate the oriented angle between two segments.

    Args:
        P1: Coordinates of the vertex
        P2: Coordinates of the first point
        P3: Coordinates of the second point

    Returns:
        The oriented angle between the two segments in degrees.
    """

    x1, y1 = P1
    x2, y2 = P2
    x3, y3 = P1
    x4, y4 = P3

    angle_AB = math.atan2(y2 - y1, x2 - x1)
    angle_CD = math.atan2(y4 - y3, x4 - x3)

    oriented_angle = math.degrees(angle_AB - angle_CD)

    return oriented_angle


def mem_info():
    """
    get info about total mem, used mem and available mem using:
       "free -m" command on Linux
       "top -l 1 -s 0" command in MacOS
       "systeminfo" command on Windows

    Returns:
        bool: True if error
        dict: values ("total_memory", "used_memory", "free_memory")
    """

    if sys.platform.startswith("linux"):
        try:
            process = subprocess.run(["free", "-m"], stdout=subprocess.PIPE)
            # out, err = process.communicate()
            out = process.stdout
            _, tot_mem, used_mem, _, _, _, available_mem = [x.decode("utf-8") for x in out.split(b"\n")[1].split(b" ") if x != b""]
            return False, {
                "total_memory": int(tot_mem),
                "used_memory": int(used_mem),
                "free_memory": int(available_mem),
            }
        except Exception:
            return True, {"msg": error_info(sys.exc_info())[0]}

    if sys.platform.startswith("darwin"):
        try:
            output = subprocess.check_output(("top", "-l", "1", "-s", "0"))
            r = [x.decode("utf-8") for x in output.split(b"\n") if b"PhysMem" in x][0].split(" ")
            used_mem, free_mem = int(r[1].replace("M", "")), int(r[5].replace("M", ""))
            return False, {"total_memory": used_mem + free_mem, "used_memory": used_mem, "free_memory": free_mem}
        except Exception:
            return True, {"msg": error_info(sys.exc_info())[0]}

    if sys.platform.startswith("win"):
        try:
            output = subprocess.run(["wmic", "computersystem", "get", "TotalPhysicalMemory", "/", "Value"], stdout=subprocess.PIPE)
            tot_mem = int(output.stdout.strip().split(b"=")[-1].decode("utf-8")) / 1024 / 1024

            output = subprocess.run(["wmic", "OS", "get", "FreePhysicalMemory", "/", "Value"], stdout=subprocess.PIPE)
            free_mem = int(output.stdout.strip().split(b"=")[-1].decode("utf-8")) / 1024
            return False, {"total_memory": tot_mem, "free_memory": free_mem}

        except Exception:
            return True, {"msg": error_info(sys.exc_info())[0]}

    return True, {"msg": "Unknown operating system"}


def polygon_area(poly: list) -> float:
    """
    area of polygon
    from http://www.mathopenref.com/coordpolygonarea.html
    """
    tot = 0
    for p in range(len(poly)):
        x1, y1 = poly[p]
        n = (p + 1) % len(poly)
        x2, y2 = poly[n]
        tot += x1 * y2 - x2 * y1

    return abs(tot / 2)


def polyline_length(poly: list) -> float:
    """
    length of polyline
    """
    tot = 0
    for p in range(1, len(poly)):
        x1, y1 = poly[p - 1]
        x2, y2 = poly[p]
        tot += ((x1 - x2) ** 2 + (y1 - y2) ** 2) ** 0.5

    return tot


def url2path(url: str) -> str:
    """
    convert URL in local path name
    under windows, check if path name begin with /
    """

    path = urllib.parse.unquote(urllib.parse.urlparse(url).path)
    # check / for windows
    if sys.platform.startswith("win") and path.startswith("/"):
        path = path[1:]
    return path


def float2decimal(f):
    """
    return decimal value
    """
    return dec(str(f))


def time2seconds(time_: str) -> dec:
    """
    convert hh:mm:ss.s to number of seconds (decimal)

    Args
        time (str): time (hh:mm:ss.zzz format)

    Returns:
        Decimal: time in seconds
    """

    if " " in time_:
        try:
            return dec(str(dt.datetime.strptime(time_, "%Y-%m-%d %H:%M:%S.%f").timestamp()))
        except Exception:
            return dec("0.000")
    else:
        try:
            flag_neg = "-" in time_
            time_ = time_.replace("-", "")
            tsplit = time_.split(":")
            h, m, s = int(tsplit[0]), int(tsplit[1]), dec(tsplit[2])
            return dec(-(h * 3600 + m * 60 + s)) if flag_neg else dec(h * 3600 + m * 60 + s)
        except Exception:
            return dec("0.000")


def seconds2time(sec: dec) -> str:
    """
    convert seconds to hh:mm:ss.sss format

    Args:
        sec (Decimal): time in seconds
    Returns:
        str: time in format hh:mm:ss
    """

    if math.isnan(sec):
        return cfg.NA

    # if sec > one day treat as date
    if sec > cfg.DATE_CUTOFF:
        t = dt.datetime.fromtimestamp(float(sec))
        return f"{t:%Y-%m-%d %H:%M:%S}.{t.microsecond / 1000:03.0f}"

    neg_sign = "-" * (sec < 0)
    abs_sec = abs(sec)

    hours = 0

    minutes = int(abs_sec / 60)
    if minutes >= 60:
        hours = int(minutes / 60)
        minutes = minutes % 60

    ssecs = f"{abs_sec - hours * 3600 - minutes * 60:06.3f}"

    return f"{neg_sign}{hours:02}:{minutes:02}:{ssecs}"


def safeFileName(s: str) -> str:
    """
    replace characters not allowed in file name by _
    """
    fileName = s
    notAllowedChars = ["/", "\\", ":", "*", "?", '"', "<", ">", "|", "\n", "\r"]
    for char in notAllowedChars:
        fileName = fileName.replace(char, "_")

    return fileName


def safe_xl_worksheet_title(title: str, output_format: str):
    """
    sanitize the XLS and XLSX worksheet title

    Args:
        title (str): title for worksheet
        output_format (str): xls or xlsx
    """
    if output_format in ("xls", "xlsx"):
        if output_format in ("xls"):
            title = title[:31]
        for forbidden_char in cfg.EXCEL_FORBIDDEN_CHARACTERS:
            title = title.replace(forbidden_char, " ")
    return title


def eol2space(s: str) -> str:
    """
    replace EOL char by space for all platforms

    Args:
        s (str): string to be converted

    Returns:
        str: string where /rn /r /n are converted in space
    """
    return s.replace("\r\n", " ").replace("\n", " ").replace("\r", " ")


def test_ffmpeg_path(FFmpegPath: str) -> Tuple[bool, str]:
    """
    test if ffmpeg has valid path

    Args:
        FFmpegPath (str): ffmepg path to test

    Returns:
        bool: True: path found
        str: message
    """

    out, error = subprocess.Popen(f'"{FFmpegPath}" -version', stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True).communicate()
    logging.debug(f"test ffmpeg path output: {out}")
    logging.debug(f"test ffmpeg path error: {error}")

    if (b"avconv" in out) or (b"the Libav developers" in error):
        return False, "Please use FFmpeg from https://www.ffmpeg.org in place of FFmpeg from Libav project."

    if (b"ffmpeg version" not in out) and (b"ffmpeg version" not in error):
        return False, "FFmpeg is required but it was not found.<br>See https://www.ffmpeg.org"

    return True, ""


def check_ffmpeg_path() -> Tuple[bool, str]:
    """
    check for ffmpeg path
    firstly search for embedded version
    if not found search for system wide version (must be in the path)

    Returns:
        bool: True if ffmpeg path found else False
        str: if bool True returns ffmpegpath else returns error message
    """

    if sys.platform.startswith("linux") or sys.platform.startswith("darwin"):
        ffmpeg_path = pl.Path("")
        # search embedded ffmpeg
        if sys.argv[0].endswith("start_boris.py"):
            ffmpeg_path = pl.Path(sys.argv[0]).resolve().parent / "boris" / "misc" / "ffmpeg"
        if sys.argv[0].endswith("__main__.py"):
            ffmpeg_path = pl.Path(sys.argv[0]).resolve().parent / "misc" / "ffmpeg"

        if not ffmpeg_path.is_file():
            # search global ffmpeg
            ffmpeg_path = "ffmpeg"

        # test ffmpeg
        r, msg = test_ffmpeg_path(str(ffmpeg_path))
        if r:
            return True, str(ffmpeg_path)
        else:
            return False, "FFmpeg is not available"

    if sys.platform.startswith("win"):
        ffmpeg_path = pl.Path("")
        # search embedded ffmpeg
        if sys.argv[0].endswith("start_boris.py"):
            ffmpeg_path = pl.Path(sys.argv[0]).resolve().parent / "boris" / "misc" / "ffmpeg.exe"
        if sys.argv[0].endswith("__main__.py"):
            ffmpeg_path = pl.Path(sys.argv[0]).resolve().parent / "misc" / "ffmpeg.exe"

        if not ffmpeg_path.is_file():
            # search global ffmpeg
            ffmpeg_path = "ffmpeg"

        # test ffmpeg
        r, msg = test_ffmpeg_path(str(ffmpeg_path))
        if r:
            return True, str(ffmpeg_path)
        else:
            return False, "FFmpeg is not available"


def smart_size_format(n: Union[float, int, str, None]) -> str:
    """
    format with kb, Mb or Gb in base of value
    """
    if n is None:
        return cfg.NA
    if str(n) == "NA":
        return cfg.NA
    if math.isnan(n):
        return cfg.NA
    if n < 1_000:
        return f"{n:,.1f} b"
    if n < 1_000_000:
        return f"{n / 1_000:,.1f} Kb"
    if n < 1_000_000_000:
        return f"{n / 1_000_000:,.1f} Mb"
    return f"{n / 1_000_000_000:,.1f} Gb"


def ffprobe_media_analysis(ffmpeg_bin: str, file_name: str) -> dict:
    """
    analyse video parameters with ffprobe (if available)

    Args:
        ffmpeg_bin (str): ffmpeg path
        file_name (str): path of media file

    Returns:
        dict
    """
    # ffprobe -v quiet -print_format json -show_format -show_streams /tmp/ramdisk/video1.mp4
    ffprobe_bin = ffmpeg_bin.replace("ffmpeg", "ffprobe")

    command = f'"{ffprobe_bin}" -hide_banner -v error -print_format json -show_format -show_streams "{file_name}"'

    p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    out, error = p.communicate()
    if error:
        return {"error": f"{error}"}

    try:
        hasVideo = False
        hasAudio = False
        """bitrate = None"""
        video_bitrate = None
        audio_bitrate = []
        resolution = None
        fps = 0
        sample_rate = None
        duration = None
        audio_duration = cfg.NA
        frames_number = None
        size = None
        audio_codec = None
        video_codec = None

        video_param = json.loads(out.decode("utf-8"))
        if "size" in video_param["format"]:
            size = int(video_param["format"]["size"])

        for stream in video_param["streams"]:
            if stream["codec_type"] == "video":
                hasVideo = True
                video_bitrate = int(stream["bit_rate"]) if "bit_rate" in stream else None
                resolution = f"{stream['width']}x{stream['height']}"

                """
                if "avg_frame_rate" in stream:
                    if stream["avg_frame_rate"] == "0/0":
                        fps = 0
                    else:
                        try:
                            fps = eval(stream["avg_frame_rate"])
                        except Exception:
                            fps = 0
                """
                if "r_frame_rate" in stream:
                    if stream["r_frame_rate"] == "0/0":
                        fps = 0
                    else:
                        try:
                            fps = eval(stream["r_frame_rate"])
                        except Exception:
                            fps = 0

                if "duration" in stream:
                    duration = float(stream["duration"])
                if "duration_ts" in stream:
                    frames_number = int(stream["duration_ts"])
                elif "nb_frames" in stream:
                    frames_number = int(stream["nb_frames"])
                else:
                    frames_number = None

                video_codec = stream["codec_long_name"] if "codec_long_name" in stream else None

            if stream["codec_type"] == "audio":
                hasAudio = True
                sample_rate = float(stream["sample_rate"]) if "sample_rate" in stream else cfg.NA
                # TODO manage audio_duration parameter
                audio_duration = float(stream["duration"]) if "duration" in stream else cfg.NA
                audio_codec = stream["codec_long_name"]
                audio_bitrate.append(int(stream.get("bit_rate", 0)))

        # check duration
        if duration is None:
            if "duration" in video_param["format"]:
                duration = float(video_param["format"]["duration"])
            else:
                duration = 0

        # check bit rate
        if "bit_rate" in video_param["format"]:
            all_bitrate = int(video_param["format"]["bit_rate"])
        else:
            all_bitrate = None

        if video_bitrate is None and all_bitrate is not None:
            video_bitrate = all_bitrate - sum(audio_bitrate)

        # extract format long name
        format_long_name = video_param["format"]["format_long_name"] if "format_long_name" in video_param["format"] else cfg.NA

        # extract creation time ("creation_time": "2023-03-22T16:50:32.000000Z")
        creation_time = cfg.NA
        if "tags" in video_param["format"] and "creation_time" in video_param["format"]["tags"]:
            creation_time = video_param["format"]["tags"]["creation_time"].replace("T", " ")
            if "." in creation_time:
                creation_time = creation_time.split(".")[0]

        return {
            "analysis_program": "ffprobe",
            "frames_number": frames_number,
            "duration_ms": duration * 1000,
            "duration": duration,
            "audio_duration": audio_duration,
            "fps": fps,
            "has_video": hasVideo,
            "has_audio": hasAudio,
            "bitrate": video_bitrate,
            "resolution": resolution,
            "sample_rate": sample_rate,
            "file size": size,
            "audio_codec": audio_codec,
            "video_codec": video_codec,
            "creation_time": creation_time,
            "format_long_name": format_long_name,
        }

    except Exception as e:
        raise
        return {"error": str(e)}


def accurate_media_analysis(ffmpeg_bin: str, file_name: str) -> dict:
    """
    analyse frame rate and video duration with ffprobe or ffmpeg if ffprobe not available
    Returns parameters: duration, duration_ms, bitrate, frames_number, fps, has_video (True/False), has_audio (True/False)

    Args:
        ffmpeg_bin (str): ffmpeg path
        file_name (str): path of media file

    Returns:
        dict containing keys: duration, duration_ms, frames_number, bitrate, fps, has_video, has_audio

    """

    ffprobe_results = ffprobe_media_analysis(ffmpeg_bin, file_name)

    logging.debug(f"file_name: {file_name}")
    logging.debug(f"ffprobe_results: {ffprobe_results}")

    if ("error" not in ffprobe_results) and (ffprobe_results["bitrate"] is not None):
        return ffprobe_results
    else:
        # use ffmpeg
        command = f'"{ffmpeg_bin}" -hide_banner -i "{file_name}" > {"NUL" if sys.platform.startswith("win") else "/dev/null"}'

        p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)

        duration, fps, hasVideo, hasAudio, bitrate = 0, 0, False, False, None
        try:
            _, error = p.communicate()
        except Exception as e:
            return {"error": str(e)}

        rows = error.split(b"\n")

        # check if file found and if invalid data found
        for row in rows:
            if b"No such file or directory" in row:
                return {"error": "No such file or directory"}
            if b"Invalid data found when processing input" in row:
                return {"error": "This file does not seem to be a media file"}

        # video duration
        try:
            for row in rows:
                if b"Duration" in row:
                    duration = time2seconds(row.split(b"Duration: ")[1].split(b",")[0].strip().decode("utf-8"))
                    break
        except Exception:
            duration = 0

        # bitrate
        try:
            for row in rows:
                if b"bitrate:" in row:
                    re_results = re.search(b"bitrate: (.{1,10}) kb", row, re.IGNORECASE)
                    if re_results:
                        bitrate = int(re_results.group(1).strip()) * 1000
                    break
        except Exception:
            bitrate = None

        # fps
        fps = 0
        try:
            for row in rows:
                if b" fps," in row:
                    re_results = re.search(b", (.{1,10}) fps,", row, re.IGNORECASE)
                    if re_results:
                        fps = dec(re_results.group(1).strip().decode("utf-8"))
                        break
        except Exception:
            fps = 0

        # check for video stream
        hasVideo, resolution = False, None
        try:
            for row in rows:
                if b"Stream #" in row and b"Video:" in row:
                    hasVideo = True
                    # get resolution \d{3,5}x\d{3,5}
                    re_results = re.search(r"\d{3,5}x\d{3,5}", row, re.IGNORECASE)
                    if re_results:
                        resolution = re_results.group(0).decode("utf-8")
                    break
        except Exception:
            hasVideo, resolution = False, None

        # check for audio stream
        hasAudio = False
        try:
            for row in rows:
                if b"Stream #" in row and b"Audio:" in row:
                    hasAudio = True
                    break
        except Exception:
            hasAudio = False

        if not hasVideo and not hasAudio:
            return {"error": "This file does not seem to be a media file"}

        return {
            "analysis_program": "ffmpeg",
            "frames_number": int(fps * duration),
            "duration_ms": duration * 1000,
            "duration": duration,
            "audio_duration": cfg.NA,
            "fps": fps,
            "has_video": hasVideo,
            "has_audio": hasAudio,
            "bitrate": bitrate,
            "resolution": resolution,
            "format_long_name": "",
        }


def behavior_color(colors_list: list, idx: int, default_color: str = "darkgray"):
    """
    return color with index corresponding to behavior index

    see BEHAVIORS_PLOT_COLORS list in config.py

    Args:
        colors_list (list): list of colors
        idx (int): index of behavior in all behaviors list (sorted)
        default_color (str): default color (if problem)

    Returns:
        str: color corresponding to behavior index

    """

    try:
        return colors_list[idx % len(colors_list)].replace("tab:", "")
    except Exception:
        return default_color


def all_behaviors(ethogram: dict) -> list:
    """
    extract all behaviors from the submitted ethogram
    behaviors are alphabetically sorted

    Args:
        ethogram (dict): ethogram

    Returns:
        list: behaviors code (alphabetically sorted)
    """

    return [ethogram[x][cfg.BEHAVIOR_CODE] for x in sorted_keys(ethogram)]


def all_subjects(subjects: dict) -> list:
    """
    extract all subjects from the subject configuration dictionary

    Args:
        subject configuration (dict)

    Returns:
        list: subjects name
    """

    return [subjects[x][cfg.SUBJECT_NAME] for x in sorted_keys(subjects)]


def dir_images_number(dir_path_str: str) -> dict:
    """
    return number of images in dir_path (see cfg.IMAGE_EXTENSIONS)
    """

    dir_path = pl.Path(dir_path_str)
    if not dir_path.is_dir():
        return {"error": f"The directory {dir_path_str} does not exists"}
    img_count = 0
    for pattern in cfg.IMAGE_EXTENSIONS:
        img_count += len(list(dir_path.glob(pattern)))
        img_count += len(list(dir_path.glob(pattern.upper())))
    return {"number of images": img_count}


def intersection(A, B, C, D):
    """
    line segments intersection with decimal precision
    return True when intersection else False
    """
    getcontext().prec = 28

    xa, ya = dec(str(A[0])), dec(str(A[1]))
    xb, yb = dec(str(B[0])), dec(str(B[1]))
    xc, yc = dec(str(C[0])), dec(str(C[1]))
    xd, yd = dec(str(D[0])), dec(str(D[1]))

    # check if first segment is vertical
    try:
        if xa == xb:
            slope = (yc - yd) / (xc - xd)
            intersept = yc - slope * xc
            xm = xa
            ym = slope * xm + intersept

        # check if second segment is vertical
        elif xc == xd:
            slope = (ya - yb) / (xa - xb)
            intersept = ya - slope * xa
            xm = xc
            ym = slope * xm + intersept
        else:
            xm = (
                (xd * xa * yc - xd * xb * yc - xd * xa * yb - xc * xa * yd + xc * xa * yb + xd * ya * xb + xc * xb * yd - xc * ya * xb)
                / (-yb * xd + yb * xc + ya * xd - ya * xc + xb * yd - xb * yc - xa * yd + xa * yc)
            ).quantize(dec(".001"), rounding=ROUND_DOWN)
            ym = (
                (yb * xc * yd - yb * yc * xd - ya * xc * yd + ya * yc * xd - xa * yb * yd + xa * yb * yc + ya * xb * yd - ya * xb * yc)
                / (-yb * xd + yb * xc + ya * xd - ya * xc + xb * yd - xb * yc - xa * yd + xa * yc)
            ).quantize(dec(".001"), rounding=ROUND_DOWN)

        xmin1, xmax1 = min(xa, xb), max(xa, xb)
        xmin2, xmax2 = min(xc, xd), max(xc, xd)
        ymin1, ymax1 = min(ya, yb), max(ya, yb)
        ymin2, ymax2 = min(yc, yd), max(yc, yd)

        return xm >= xmin1 and xm <= xmax1 and xm >= xmin2 and xm <= xmax2 and ym >= ymin1 and ym <= ymax1 and ym >= ymin2 and ym <= ymax2

    except Exception:  # for cases xa=xb=xc=xd
        return True



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'boris/param_panel.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(1037, 890)
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.lbSubjects = QtWidgets.QLabel(Dialog)
        self.lbSubjects.setObjectName("lbSubjects")
        self.verticalLayout_2.addWidget(self.lbSubjects)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.pbSelectAllSubjects = QtWidgets.QPushButton(Dialog)
        self.pbSelectAllSubjects.setObjectName("pbSelectAllSubjects")
        self.horizontalLayout_3.addWidget(self.pbSelectAllSubjects)
        self.pbUnselectAllSubjects = QtWidgets.QPushButton(Dialog)
        self.pbUnselectAllSubjects.setObjectName("pbUnselectAllSubjects")
        self.horizontalLayout_3.addWidget(self.pbUnselectAllSubjects)
        self.pbReverseSubjectsSelection = QtWidgets.QPushButton(Dialog)
        self.pbReverseSubjectsSelection.setObjectName("pbReverseSubjectsSelection")
        self.horizontalLayout_3.addWidget(self.pbReverseSubjectsSelection)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem)
        self.verticalLayout_2.addLayout(self.horizontalLayout_3)
        self.lwSubjects = QtWidgets.QListWidget(Dialog)
        self.lwSubjects.setObjectName("lwSubjects")
        self.verticalLayout_2.addWidget(self.lwSubjects)
        self.lbBehaviors = QtWidgets.QLabel(Dialog)
        self.lbBehaviors.setObjectName("lbBehaviors")
        self.verticalLayout_2.addWidget(self.lbBehaviors)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.pbSelectAllBehaviors = QtWidgets.QPushButton(Dialog)
        self.pbSelectAllBehaviors.setObjectName("pbSelectAllBehaviors")
        self.horizontalLayout_4.addWidget(self.pbSelectAllBehaviors)
        self.pbUnselectAllBehaviors = QtWidgets.QPushButton(Dialog)
        self.pbUnselectAllBehaviors.setObjectName("pbUnselectAllBehaviors")
        self.horizontalLayout_4.addWidget(self.pbUnselectAllBehaviors)
        self.pbReverseBehaviorsSelection = QtWidgets.QPushButton(Dialog)
        self.pbReverseBehaviorsSelection.setObjectName("pbReverseBehaviorsSelection")
        self.horizontalLayout_4.addWidget(self.pbReverseBehaviorsSelection)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem1)
        self.verticalLayout_2.addLayout(self.horizontalLayout_4)
        self.lwBehaviors = QtWidgets.QListWidget(Dialog)
        self.lwBehaviors.setObjectName("lwBehaviors")
        self.verticalLayout_2.addWidget(self.lwBehaviors)
        self.horizontalLayout_9 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_9.setObjectName("horizontalLayout_9")
        self.cbIncludeModifiers = QtWidgets.QCheckBox(Dialog)
        self.cbIncludeModifiers.setObjectName("cbIncludeModifiers")
        self.horizontalLayout_9.addWidget(self.cbIncludeModifiers)
        self.cbExcludeBehaviors = QtWidgets.QCheckBox(Dialog)
        self.cbExcludeBehaviors.setObjectName("cbExcludeBehaviors")
        self.horizontalLayout_9.addWidget(self.cbExcludeBehaviors)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_9.addItem(spacerItem2)
        self.verticalLayout_2.addLayout(self.horizontalLayout_9)
        self.frm_time_bin_size = QtWidgets.QFrame(Dialog)
        self.frm_time_bin_size.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frm_time_bin_size.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frm_time_bin_size.setObjectName("frm_time_bin_size")
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout(self.frm_time_bin_size)
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.lb_time_bin_size = QtWidgets.QLabel(self.frm_time_bin_size)
        self.lb_time_bin_size.setObjectName("lb_time_bin_size")
        self.horizontalLayout_7.addWidget(self.lb_time_bin_size)
        self.sb_time_bin_size = QtWidgets.QSpinBox(self.frm_time_bin_size)
        self.sb_time_bin_size.setMaximum(86400)
        self.sb_time_bin_size.setSingleStep(10)
        self.sb_time_bin_size.setObjectName("sb_time_bin_size")
        self.horizontalLayout_7.addWidget(self.sb_time_bin_size)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_7.addItem(spacerItem3)
        self.horizontalLayout_8.addLayout(self.horizontalLayout_7)
        self.verticalLayout_2.addWidget(self.frm_time_bin_size)
        self.frm_time = QtWidgets.QFrame(Dialog)
        self.frm_time.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frm_time.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frm_time.setObjectName("frm_time")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.frm_time)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.lb_time_interval = QtWidgets.QLabel(self.frm_time)
        self.lb_time_interval.setObjectName("lb_time_interval")
        self.verticalLayout_3.addWidget(self.lb_time_interval)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.rb_observed_events = QtWidgets.QRadioButton(self.frm_time)
        self.rb_observed_events.setObjectName("rb_observed_events")
        self.horizontalLayout_5.addWidget(self.rb_observed_events)
        self.rb_user_defined = QtWidgets.QRadioButton(self.frm_time)
        self.rb_user_defined.setObjectName("rb_user_defined")
        self.horizontalLayout_5.addWidget(self.rb_user_defined)
        self.rb_media_duration = QtWidgets.QRadioButton(self.frm_time)
        self.rb_media_duration.setCheckable(True)
        self.rb_media_duration.setChecked(False)
        self.rb_media_duration.setObjectName("rb_media_duration")
        self.horizontalLayout_5.addWidget(self.rb_media_duration)
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem4)
        self.verticalLayout_3.addLayout(self.horizontalLayout_5)
        self.frm_time_interval = QtWidgets.QFrame(self.frm_time)
        self.frm_time_interval.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frm_time_interval.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frm_time_interval.setObjectName("frm_time_interval")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.frm_time_interval)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.lbStartTime = QtWidgets.QLabel(self.frm_time_interval)
        self.lbStartTime.setObjectName("lbStartTime")
        self.horizontalLayout.addWidget(self.lbStartTime)
        self.label_2 = QtWidgets.QLabel(self.frm_time_interval)
        self.label_2.setText("")
        self.label_2.setObjectName("label_2")
        self.horizontalLayout.addWidget(self.label_2)
        spacerItem5 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem5)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.lbEndTime = QtWidgets.QLabel(self.frm_time_interval)
        self.lbEndTime.setObjectName("lbEndTime")
        self.horizontalLayout_6.addWidget(self.lbEndTime)
        self.label_3 = QtWidgets.QLabel(self.frm_time_interval)
        self.label_3.setText("")
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_6.addWidget(self.label_3)
        spacerItem6 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_6.addItem(spacerItem6)
        self.verticalLayout.addLayout(self.horizontalLayout_6)
        spacerItem7 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem7)
        self.verticalLayout_3.addWidget(self.frm_time_interval)
        self.verticalLayout_2.addWidget(self.frm_time)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem8 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem8)
        self.pbCancel = QtWidgets.QPushButton(Dialog)
        self.pbCancel.setObjectName("pbCancel")
        self.horizontalLayout_2.addWidget(self.pbCancel)
        self.pbOK = QtWidgets.QPushButton(Dialog)
        self.pbOK.setObjectName("pbOK")
        self.horizontalLayout_2.addWidget(self.pbOK)
        self.verticalLayout_2.addLayout(self.horizontalLayout_2)
        self.verticalLayout_4.addLayout(self.verticalLayout_2)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Parameters"))
        self.lbSubjects.setText(_translate("Dialog", "Subjects"))
        self.pbSelectAllSubjects.setText(_translate("Dialog", "Select all"))
        self.pbUnselectAllSubjects.setText(_translate("Dialog", "Unselect all"))
        self.pbReverseSubjectsSelection.setText(_translate("Dialog", "Reverse selection"))
        self.lbBehaviors.setText(_translate("Dialog", "Behaviors"))
        self.pbSelectAllBehaviors.setText(_translate("Dialog", "Select all"))
        self.pbUnselectAllBehaviors.setText(_translate("Dialog", "Unselect all"))
        self.pbReverseBehaviorsSelection.setText(_translate("Dialog", "Reverse selection"))
        self.cbIncludeModifiers.setText(_translate("Dialog", "Include modifiers"))
        self.cbExcludeBehaviors.setText(_translate("Dialog", "Exclude behaviors without events"))
        self.lb_time_bin_size.setText(_translate("Dialog", "Time bin size (s)"))
        self.lb_time_interval.setText(_translate("Dialog", "Time interval"))
        self.rb_observed_events.setText(_translate("Dialog", "Observed events"))
        self.rb_user_defined.setText(_translate("Dialog", "User defined"))
        self.rb_media_duration.setText(_translate("Dialog", "Media file(s) duration"))
        self.lbStartTime.setText(_translate("Dialog", "Start time"))
        self.lbEndTime.setText(_translate("Dialog", "End time"))
        self.pbCancel.setText(_translate("Dialog", "Cancel"))
        self.pbOK.setText(_translate("Dialog", "OK"))



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import json
import logging
import re

from PyQt5.QtCore import Qt, QDateTime
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import (
    QCheckBox,
    QDialog,
    QFileDialog,
    QHBoxLayout,
    QHeaderView,
    QInputDialog,
    QLabel,
    QLineEdit,
    QMenu,
    QMessageBox,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QTableWidgetItem,
    QVBoxLayout,
    QColorDialog,
    QTableWidget,
    QAbstractItemView,
)

from . import add_modifier
from . import config as cfg
from . import utilities as util
from . import converters, dialog, exclusion_matrix, project_import_export
from .project_ui import Ui_dlgProject


class BehavioralCategories(QDialog):
    """
    Class for managing the behavioral categories
    """

    def __init__(self, pj):
        super().__init__()

        self.pj = pj
        self.setWindowTitle("Behavioral categories")

        self.renamed = None
        self.removed = None

        self.vbox = QVBoxLayout(self)

        self.label = QLabel()
        self.label.setText("Behavioral categories")
        self.vbox.addWidget(self.label)

        # self.lw = QListWidget()
        self.lw = QTableWidget()
        self.lw.cellDoubleClicked[int, int].connect(self.lw_double_clicked)

        # add categories
        self.lw.setColumnCount(2)
        self.lw.setHorizontalHeaderLabels(["Category name", "Color"])
        # self.lw.verticalHeader().hide()
        self.lw.setEditTriggers(QAbstractItemView.NoEditTriggers)

        # self.lw.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.lw.setSelectionMode(QAbstractItemView.SingleSelection)

        if cfg.BEHAVIORAL_CATEGORIES_CONF in pj:
            self.lw.setRowCount(len(pj.get(cfg.BEHAVIORAL_CATEGORIES_CONF, {})))
            behav_cat = pj.get(cfg.BEHAVIORAL_CATEGORIES_CONF, {})
            for idx, key in enumerate(behav_cat.keys()):
                # name
                item = QTableWidgetItem()
                item.setText(behav_cat[key]["name"])
                # item.setFlags(Qt.ItemIsEnabled)
                self.lw.setItem(idx, 0, item)
                # color
                item = QTableWidgetItem()
                item.setText(behav_cat[key].get(cfg.COLOR, ""))
                if behav_cat[key].get(cfg.COLOR, ""):
                    item.setBackground(QColor(behav_cat[key].get(cfg.COLOR, "")))
                else:
                    item.setBackground(QColor(230, 230, 230))
                # item.setFlags(Qt.ItemIsEnabled)
                self.lw.setItem(idx, 1, item)
        else:
            self.lw.setRowCount(len(pj.get(cfg.BEHAVIORAL_CATEGORIES, [])))
            for idx, category in enumerate(sorted(pj.get(cfg.BEHAVIORAL_CATEGORIES, []))):
                item = QTableWidgetItem()
                item.setText(category)
                # item.setFlags(Qt.ItemIsEnabled)
                self.lw.setItem(idx, 0, item)

                item = QTableWidgetItem()
                item.setText("")
                # item.setFlags(Qt.ItemIsEnabled)
                self.lw.setItem(idx, 1, item)

            # self.lw.addItem(QListWidgetItem(category))

        self.vbox.addWidget(self.lw)

        self.hbox0 = QHBoxLayout()
        self.pbAddCategory = QPushButton("Add category", clicked=self.add_behavioral_category)
        self.pbRemoveCategory = QPushButton("Remove category", clicked=self.remove_behavioral_category)
        self.pb_rename_category = QPushButton("Rename category", clicked=self.pb_rename_category_clicked)

        spacerItem = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.hbox0.addItem(spacerItem)
        self.hbox0.addWidget(self.pb_rename_category)
        self.hbox0.addWidget(self.pbRemoveCategory)
        self.hbox0.addWidget(self.pbAddCategory)
        self.vbox.addLayout(self.hbox0)

        hbox1 = QHBoxLayout()
        self.pbOK = QPushButton("OK", clicked=self.accept)
        self.pbCancel = QPushButton("Cancel", clicked=self.accept)

        spacerItem = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        hbox1.addItem(spacerItem)
        hbox1.addWidget(self.pbCancel)
        hbox1.addWidget(self.pbOK)
        self.vbox.addLayout(hbox1)

        self.setLayout(self.vbox)

    def lw_double_clicked(self, row: int, column: int):
        """
        change color
        """

        if column != 1:
            return
        col_diag = QColorDialog()
        col_diag.setOptions(QColorDialog.DontUseNativeDialog)

        if self.lw.item(row, 1).text():
            current_color = QColor(self.lw.item(row, 1).text())
            if current_color.isValid():
                col_diag.setCurrentColor(current_color)

        if col_diag.exec_():
            color = col_diag.currentColor()
            if color.name() == "#000000":  # black -> delete color
                self.lw.item(row, 1).setText("")
                self.lw.item(row, 1).setBackground(QColor(230, 230, 230))
            elif color.isValid():
                self.lw.item(row, 1).setText(color.name())
                self.lw.item(row, 1).setBackground(color)

    def add_behavioral_category(self):
        """
        add a behavioral category
        """
        category, ok = QInputDialog.getText(self, "New behavioral category", "Category name:")
        if ok:
            self.lw.insertRow(self.lw.rowCount())
            item = QTableWidgetItem(category)
            self.lw.setItem(self.lw.rowCount() - 1, 0, item)

            item = QTableWidgetItem("")
            # item.setFlags(Qt.ItemIsEnabled)
            self.lw.setItem(self.lw.rowCount() - 1, 1, item)

    def remove_behavioral_category(self):
        """
        remove the selected behavioral category
        """

        for selected_item in self.lw.selectedItems():
            # check if behavioral category is in use
            if (
                dialog.MessageDialog(
                    cfg.programName,
                    ("Confirm deletion of the behavioral category"),
                    ("Confirm", cfg.CANCEL),
                )
                == cfg.CANCEL
            ):
                continue

            category_to_remove = self.lw.item(self.lw.row(selected_item), 0).text().strip()
            behaviors_in_category: list = []
            for idx in self.pj[cfg.ETHOGRAM]:
                if (
                    cfg.BEHAVIOR_CATEGORY in self.pj[cfg.ETHOGRAM][idx]
                    and self.pj[cfg.ETHOGRAM][idx][cfg.BEHAVIOR_CATEGORY] == category_to_remove
                ):
                    behaviors_in_category.append(self.pj[cfg.ETHOGRAM][idx][cfg.BEHAVIOR_CODE])
            flag_remove = False
            if behaviors_in_category:
                flag_remove = (
                    dialog.MessageDialog(
                        cfg.programName,
                        (
                            f"Some behavior belong to the <b>{category_to_remove}</b> to remove:<br>"
                            f"{'<br>'.join(behaviors_in_category)}<br>"
                            "<br>Some features may not be available anymore.<br>"
                        ),
                        ("Remove category", cfg.CANCEL),
                    )
                    == "Remove category"
                )

            else:
                flag_remove = True

            if flag_remove:
                self.lw.removeRow(self.lw.row(selected_item))
                self.removed = category_to_remove

                self.accept()

    def pb_rename_category_clicked(self, row: int):
        """
        rename the selected behavioral category
        """
        for selected_item in self.lw.selectedItems():
            # check if behavioral category is in use
            category_to_rename = self.lw.item(self.lw.row(selected_item), 0).text().strip()
            behaviors_in_category = []
            for idx in self.pj[cfg.ETHOGRAM]:
                if (
                    cfg.BEHAVIOR_CATEGORY in self.pj[cfg.ETHOGRAM][idx]
                    and self.pj[cfg.ETHOGRAM][idx][cfg.BEHAVIOR_CATEGORY] == category_to_rename
                ):
                    behaviors_in_category.append(self.pj[cfg.ETHOGRAM][idx][cfg.BEHAVIOR_CODE])

            flag_rename = False
            if behaviors_in_category:
                flag_rename = (
                    dialog.MessageDialog(
                        cfg.programName,
                        ("Some behavior belong to the <b>{1}</b> to rename:<br>" "{0}<br>").format(
                            "<br>".join(behaviors_in_category), category_to_rename
                        ),
                        ["Rename category", cfg.CANCEL],
                    )
                    == "Rename category"
                )
            else:
                flag_rename = True

            if flag_rename:
                new_category_name, ok = QInputDialog.getText(
                    self, "Rename behavioral category", "New category name:", QLineEdit.Normal, category_to_rename
                )
                if ok:
                    self.lw.item(self.lw.indexFromItem(selected_item).row(), 0).setText(new_category_name)
                    # check behaviors belonging to the renamed category
                    self.renamed = [category_to_rename, new_category_name]
                    self.accept()


class projectDialog(QDialog, Ui_dlgProject):
    def __init__(self, parent=None):
        super().__init__()

        self.setupUi(self)

        self.lbObservationsState.setText("")
        self.lbSubjectsState.setText("")

        # ethogram tab
        behavior_button_items = [
            "new|Add new behavior",
            "clone|Clone behavior",
            "remove|Remove behavior",
            "remove all|Remove all behaviors",
            "lower|Convert keys to lower case",
        ]
        menu = QMenu()
        menu.triggered.connect(lambda x: self.behavior(action=x.statusTip()))
        self.add_button_menu(behavior_button_items, menu)
        self.pb_behavior.setMenu(menu)

        import_button_items = [
            "boris|from a BORIS project",
            "spreadsheet|from a spreadsheet file (XLSX/ODS)",
            "jwatcher|from a JWatcher project",
            "text|from a text file (CSV or TSV)",
            "clipboard|from the clipboard",
            "repository|from the BORIS repository",
        ]
        menu = QMenu()
        menu.triggered.connect(lambda x: self.import_ethogram(action=x.statusTip()))
        self.add_button_menu(import_button_items, menu)
        self.pb_import.setMenu(menu)

        self.pbBehaviorsCategories.clicked.connect(self.pbBehaviorsCategories_clicked)

        self.pb_exclusion_matrix.clicked.connect(self.exclusion_matrix)

        self.pbExportEthogram.clicked.connect(lambda: project_import_export.export_ethogram(self))

        self.twBehaviors.cellChanged[int, int].connect(self.twBehaviors_cellChanged)
        self.twBehaviors.cellDoubleClicked[int, int].connect(self.twBehaviors_cellDoubleClicked)

        # left align table header
        for i in range(self.twBehaviors.columnCount()):
            self.twBehaviors.horizontalHeaderItem(i).setTextAlignment(Qt.AlignLeft)

        # subjects
        subjects_button_items = [
            "new|Add a new subject",
            # "clone|Clone subject",
            "remove|Remove subject",
            "remove all|Remove all subjects",
            "lower|Convert keys to lower case",
        ]

        menu = QMenu()
        menu.triggered.connect(lambda x: self.subjects(action=x.statusTip()))
        self.add_button_menu(subjects_button_items, menu)
        self.pb_subjects.setMenu(menu)

        subjects_import_button_items = [
            "boris|from a BORIS project",
            "spreadsheet|from a spreadsheet file (XLSX/ODS)",
            "text|from a text file (CSV or TSV)",
            "clipboard|from the clipboard",
        ]
        menu = QMenu()
        menu.triggered.connect(lambda x: self.import_subjects(action=x.statusTip()))
        self.add_button_menu(subjects_import_button_items, menu)
        self.pbImportSubjectsFromProject.setMenu(menu)

        self.pb_export_subjects.clicked.connect(lambda: project_import_export.export_subjects(self))

        self.twSubjects.cellChanged[int, int].connect(self.twSubjects_cellChanged)

        # independent variables tab
        self.pbAddVariable.clicked.connect(self.pbAddVariable_clicked)
        self.pbRemoveVariable.clicked.connect(self.pbRemoveVariable_clicked)

        self.leLabel.textChanged.connect(self.leLabel_changed)
        self.leDescription.textChanged.connect(self.leDescription_changed)
        self.lePredefined.textChanged.connect(self.lePredefined_changed)
        self.leSetValues.textChanged.connect(self.leSetValues_changed)
        self.dte_default_date.dateTimeChanged.connect(self.dte_default_date_changed)

        self.twVariables.cellClicked[int, int].connect(self.twVariables_cellClicked)

        self.cbType.currentIndexChanged.connect(self.cbtype_changed)
        self.cbType.activated.connect(self.cbtype_activated)

        self.pbImportVarFromProject.clicked.connect(self.pbImportVarFromProject_clicked)

        self.pbOK.clicked.connect(self.pbOK_clicked)
        self.pbCancel.clicked.connect(self.pbCancel_clicked)

        self.selected_twvariables_row = -1

        self.pbAddBehaviorsCodingMap.clicked.connect(self.add_behaviors_coding_map)
        self.pbRemoveBehaviorsCodingMap.clicked.connect(self.remove_behaviors_coding_map)

        # converters tab
        self.pb_add_converter.clicked.connect(lambda: converters.add_converter(self))
        self.pb_modify_converter.clicked.connect(lambda: converters.modify_converter(self))
        self.pb_save_converter.clicked.connect(lambda: converters.save_converter(self))
        self.pb_cancel_converter.clicked.connect(lambda: converters.cancel_converter(self))
        self.pb_delete_converter.clicked.connect(lambda: converters.delete_converter(self))

        self.pb_load_from_file.clicked.connect(lambda: converters.load_converters_from_file_repo(self, mode="file"))
        self.pb_load_from_repo.clicked.connect(lambda: converters.load_converters_from_file_repo(self, mode="repo"))

        self.pb_code_help.clicked.connect(lambda: converters.pb_code_help_clicked(self))

        self.row_in_modification = -1
        self.flag_modified = False

        for w in (
            self.le_converter_name,
            self.le_converter_description,
            self.pteCode,
            self.pb_save_converter,
            self.pb_cancel_converter,
        ):
            w.setEnabled(False)

        # disable widget for indep var setting
        for widget in (
            self.leLabel,
            self.le_converter_description,
            self.cbType,
            self.lePredefined,
            self.dte_default_date,
            self.leSetValues,
        ):
            widget.setEnabled(False)

        self.twBehaviors.horizontalHeader().sortIndicatorChanged.connect(self.sort_twBehaviors)
        self.twSubjects.horizontalHeader().sortIndicatorChanged.connect(self.sort_twSubjects)
        self.twVariables.horizontalHeader().sortIndicatorChanged.connect(self.sort_twVariables)

    def add_button_menu(self, data, menu_obj):
        """
        add menu option from dictionary
        """
        if isinstance(data, dict):
            for k, v in data.items():
                sub_menu = QMenu(k, menu_obj)
                menu_obj.addMenu(sub_menu)
                self.add_button_menu(v, sub_menu)
        elif isinstance(data, list):
            for element in data:
                self.add_button_menu(element, menu_obj)
        else:
            action = menu_obj.addAction(data.split("|")[1])
            # tips are used to discriminate the menu option
            action.setStatusTip(data.split("|")[0])
            action.setIconVisibleInMenu(False)

    def behavior(self, action: str):
        """
        behavior menu
        """
        if action == "new":
            self.add_behavior()
        if action == "clone":
            self.clone_behavior()
        if action == "remove":
            self.remove_behavior()
        if action == "remove all":
            self.remove_all_behaviors()
        if action == "lower":
            self.convert_behaviors_keys_to_lower_case()

    def import_ethogram(self, action: str):
        """
        import behaviors
        """
        if action == "boris":
            project_import_export.import_behaviors_from_project(self)
        if action == "jwatcher":
            project_import_export.import_behaviors_from_JWatcher(self)
        if action == "text":
            project_import_export.import_behaviors_from_text_file(self)
        if action == "spreadsheet":
            project_import_export.import_behaviors_from_spreadsheet(self)
        if action == "clipboard":
            project_import_export.import_behaviors_from_clipboard(self)
        if action == "repository":
            project_import_export.import_behaviors_from_repository(self)

    def import_subjects(self, action: str):
        """
        import subjects
        """
        if action == "boris":
            project_import_export.import_subjects_from_project(self)
        if action == "text":
            project_import_export.import_subjects_from_text_file(self)
        if action == "spreadsheet":
            project_import_export.import_subjects_from_spreadsheet(self)
        if action == "clipboard":
            project_import_export.import_subjects_from_clipboard(self)

    def subjects(self, action: str):
        """
        subjects menu
        """
        if action == "new":
            self.add_subject()
        # if action == "clone":
        #    self.clone_behavior()
        if action == "remove":
            self.remove_subject()
        if action == "remove all":
            self.remove_all_subjects()
        if action == "lower":
            self.convert_subjects_keys_to_lower_case()

    def sort_twBehaviors(self, index, order):
        """
        order ethogram table
        """
        self.twBehaviors.sortItems(index, order)

    def sort_twSubjects(self, index, order):
        """
        order subjects table
        """
        self.twSubjects.sortItems(index, order)

    def sort_twVariables(self, index, order):
        """
        order variables table
        """
        self.twVariables.sortItems(index, order)

    def convert_behaviors_keys_to_lower_case(self):
        """
        convert behaviors key to lower case to help to migrate to v. 7
        """

        if not self.twBehaviors.rowCount():
            QMessageBox.critical(
                None,
                cfg.programName,
                "The ethogram is empty",
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )
            return

        # check if some keys will be duplicated after conversion
        try:
            all_keys = [self.twBehaviors.item(row, cfg.behavioursFields["key"]).text() for row in range(self.twBehaviors.rowCount())]
        except Exception:
            pass
        if all_keys == [x.lower() for x in all_keys]:
            QMessageBox.information(self, cfg.programName, "All keys are already lower case")
            return

        if dialog.MessageDialog(cfg.programName, "Confirm the conversion of key to lower case.", [cfg.YES, cfg.CANCEL]) == cfg.CANCEL:
            return

        if len([x.lower() for x in all_keys]) != len(set([x.lower() for x in all_keys])):
            if (
                dialog.MessageDialog(
                    cfg.programName,
                    "Some keys will be duplicated after conversion. Proceed?",
                    [cfg.YES, cfg.CANCEL],
                )
                == cfg.CANCEL
            ):
                return

        for row in range(self.twBehaviors.rowCount()):
            if self.twBehaviors.item(row, cfg.behavioursFields["key"]).text():
                self.twBehaviors.item(row, cfg.behavioursFields["key"]).setText(
                    self.twBehaviors.item(row, cfg.behavioursFields["key"]).text().lower()
                )

            # convert modifier shortcuts
            if self.twBehaviors.item(row, cfg.behavioursFields[cfg.MODIFIERS]).text():
                modifiers_dict = (
                    eval(self.twBehaviors.item(row, cfg.behavioursFields[cfg.MODIFIERS]).text())
                    if self.twBehaviors.item(row, cfg.behavioursFields[cfg.MODIFIERS]).text()
                    else {}
                )

                for modifier_set in modifiers_dict:
                    try:
                        for idx2, value in enumerate(modifiers_dict[modifier_set]["values"]):
                            if re.findall(r"\((\w+)\)", value):
                                modifiers_dict[modifier_set]["values"][idx2] = (
                                    value.split("(")[0] + "(" + re.findall(r"\((\w+)\)", value)[0].lower() + ")" + value.split(")")[-1]
                                )
                    except Exception:
                        logging.warning("error during conversion of modifier short cut to lower case")

                self.twBehaviors.item(row, cfg.behavioursFields[cfg.MODIFIERS]).setText(str(modifiers_dict))

    def convert_subjects_keys_to_lower_case(self):
        """
        convert subjects key to lower case to help to migrate to v. 7
        """
        # check if some keys will be duplicated after conversion
        try:
            all_keys = [self.twSubjects.item(row, cfg.subjectsFields.index("key")).text() for row in range(self.twSubjects.rowCount())]
        except Exception:
            pass
        if all_keys == [x.lower() for x in all_keys]:
            QMessageBox.information(self, cfg.programName, "All keys are already lower case")
            return

        if dialog.MessageDialog(cfg.programName, "Confirm the conversion of key to lower case.", [cfg.YES, cfg.CANCEL]) == cfg.CANCEL:
            return

        if len([x.lower() for x in all_keys]) != len(set([x.lower() for x in all_keys])):
            if (
                dialog.MessageDialog(
                    cfg.programName,
                    "Some keys will be duplicated after conversion. Proceed?",
                    [cfg.YES, cfg.CANCEL],
                )
                == cfg.CANCEL
            ):
                return

        for row in range(self.twSubjects.rowCount()):
            if self.twSubjects.item(row, cfg.subjectsFields.index("key")).text():
                self.twSubjects.item(row, cfg.subjectsFields.index("key")).setText(
                    self.twSubjects.item(row, cfg.subjectsFields.index("key")).text().lower()
                )

    def add_behaviors_coding_map(self):
        """
        Add a behaviors coding map from file
        """

        fn = QFileDialog().getOpenFileName(
            self, "Open a behaviors coding map", "", "Behaviors coding map (*.behav_coding_map);;All files (*)"
        )
        file_name = fn[0] if type(fn) is tuple else fn
        if file_name:
            try:
                bcm = json.loads(open(file_name, "r").read())
            except Exception:
                QMessageBox.critical(self, cfg.programName, f"The file {file_name} is not a behaviors coding map.")
                return

            if "coding_map_type" not in bcm or bcm["coding_map_type"] != "BORIS behaviors coding map":
                QMessageBox.critical(self, cfg.programName, f"The file {file_name} is not a BORIS behaviors coding map.")

            if cfg.BEHAVIORS_CODING_MAP not in self.pj:
                self.pj[cfg.BEHAVIORS_CODING_MAP] = []

            bcm_code_not_found = []
            existing_codes = [self.pj[cfg.ETHOGRAM][key][cfg.BEHAVIOR_CODE] for key in self.pj[cfg.ETHOGRAM]]
            for code in [bcm["areas"][key][cfg.BEHAVIOR_CODE] for key in bcm["areas"]]:
                if code not in existing_codes:
                    bcm_code_not_found.append(code)

            if bcm_code_not_found:
                QMessageBox.warning(
                    self,
                    cfg.programName,
                    ("The following behavior{} are not defined in the ethogram:<br>" "{}").format(
                        "s" if len(bcm_code_not_found) > 1 else "", ",".join(bcm_code_not_found)
                    ),
                )

            self.pj[cfg.BEHAVIORS_CODING_MAP].append(dict(bcm))

            self.twBehavCodingMap.setRowCount(self.twBehavCodingMap.rowCount() + 1)

            self.twBehavCodingMap.setItem(self.twBehavCodingMap.rowCount() - 1, 0, QTableWidgetItem(bcm["name"]))
            codes = ", ".join([bcm["areas"][idx][cfg.BEHAVIOR_CODE] for idx in bcm["areas"]])
            self.twBehavCodingMap.setItem(self.twBehavCodingMap.rowCount() - 1, 1, QTableWidgetItem(codes))

    def remove_behaviors_coding_map(self):
        """
        remove the first selected behaviors coding map
        """
        if not self.twBehavCodingMap.selectedIndexes():
            QMessageBox.warning(self, cfg.programName, "Select a behaviors coding map")
        else:
            if dialog.MessageDialog(cfg.programName, "Remove the selected behaviors coding map?", [cfg.YES, cfg.CANCEL]) == cfg.YES:
                del self.pj[cfg.BEHAVIORS_CODING_MAP][self.twBehavCodingMap.selectedIndexes()[0].row()]
                self.twBehavCodingMap.removeRow(self.twBehavCodingMap.selectedIndexes()[0].row())

    def leLabel_changed(self):
        if self.selected_twvariables_row != -1:
            self.twVariables.item(self.selected_twvariables_row, 0).setText(self.leLabel.text())

    def leDescription_changed(self):
        if self.selected_twvariables_row != -1:
            self.twVariables.item(self.selected_twvariables_row, 1).setText(self.leDescription.text())

    def lePredefined_changed(self):
        if self.selected_twvariables_row != -1:
            self.twVariables.item(self.selected_twvariables_row, 3).setText(self.lePredefined.text())
            if not self.lePredefined.hasFocus():
                r, msg = self.check_indep_var_config()
                if not r:
                    QMessageBox.warning(self, f"{cfg.programName} - Independent variables error", msg)

    def leSetValues_changed(self):
        if self.selected_twvariables_row != -1:
            self.twVariables.item(self.selected_twvariables_row, 4).setText(self.leSetValues.text())

    def dte_default_date_changed(self):
        """
        default timestamp changed
        """
        if self.selected_twvariables_row != -1:
            self.twVariables.item(self.selected_twvariables_row, 3).setText(
                self.dte_default_date.dateTime().toString("yyyy-MM-ddTHH:mm:ss.zzz")
            )

    def pbBehaviorsCategories_clicked(self):
        """
        behavioral categories manager
        """

        bc = BehavioralCategories(self.pj)

        if bc.exec_():
            self.pj[cfg.BEHAVIORAL_CATEGORIES] = []
            self.pj[cfg.BEHAVIORAL_CATEGORIES_CONF] = {}
            for index in range(bc.lw.rowCount()):
                self.pj[cfg.BEHAVIORAL_CATEGORIES].append(bc.lw.item(index, 0).text().strip())
                self.pj[cfg.BEHAVIORAL_CATEGORIES_CONF][str(index)] = {
                    "name": bc.lw.item(index, 0).text().strip(),
                    cfg.COLOR: bc.lw.item(index, 1).text(),
                }

            # sort
            self.pj[cfg.BEHAVIORAL_CATEGORIES] = sorted(self.pj[cfg.BEHAVIORAL_CATEGORIES])

            # check if behavior belong to removed category
            if bc.removed:
                for row in range(self.twBehaviors.rowCount()):
                    if self.twBehaviors.item(row, cfg.behavioursFields[cfg.BEHAVIOR_CATEGORY]):
                        if self.twBehaviors.item(row, cfg.behavioursFields[cfg.BEHAVIOR_CATEGORY]).text() == bc.removed:
                            if (
                                dialog.MessageDialog(
                                    cfg.programName,
                                    (
                                        f"The <b>{self.twBehaviors.item(row, cfg.behavioursFields['code']).text()}</b> behavior belongs "
                                        "to a behavioral category "
                                        f"<b>{self.twBehaviors.item(row, cfg.behavioursFields['category']).text()}</b> "
                                        "that is no more in the behavioral categories list.<br><br>"
                                        "Remove the behavior from category?"
                                    ),
                                    [cfg.YES, cfg.CANCEL],
                                )
                                == cfg.YES
                            ):
                                self.twBehaviors.item(row, cfg.behavioursFields[cfg.BEHAVIOR_CATEGORY]).setText("")
            if bc.renamed:
                for row in range(self.twBehaviors.rowCount()):
                    if self.twBehaviors.item(row, cfg.behavioursFields[cfg.BEHAVIOR_CATEGORY]):
                        if self.twBehaviors.item(row, cfg.behavioursFields[cfg.BEHAVIOR_CATEGORY]).text() == bc.renamed[0]:
                            self.twBehaviors.item(row, cfg.behavioursFields[cfg.BEHAVIOR_CATEGORY]).setText(bc.renamed[1])

    def twBehaviors_cellDoubleClicked(self, row: int, column: int) -> None:
        """
        manage double-click on ethogram table:
        * color
        * behavioral category
        * modifiers
        * exclusion
        * modifiers coding map

        Args:
            row (int): row double-clicked
            column (int): column double-clicked
        """

        # check if double click on excluded column
        if column == cfg.behavioursFields["excluded"]:
            self.exclusion_matrix()

        # check if double click on 'coding map' column
        if column == cfg.behavioursFields["coding map"]:
            if "with coding map" in self.twBehaviors.item(row, cfg.behavioursFields[cfg.TYPE]).text():
                self.behaviorTypeChanged(row)
            else:
                QMessageBox.information(self, cfg.programName, "Change the behavior type on first column to select a coding map")

        # check if double click on category
        if column == cfg.behavioursFields["type"]:
            self.behavior_type_doubleclicked(row)

        # color
        if column == cfg.behavioursFields[cfg.COLOR]:
            self.color_doubleclicked(row)

        # behavioral category
        if column == cfg.behavioursFields[cfg.BEHAVIOR_CATEGORY]:
            self.category_doubleclicked(row)

        if column == cfg.behavioursFields[cfg.MODIFIERS]:
            # check if behavior has coding map
            if (
                self.twBehaviors.item(row, cfg.behavioursFields["coding map"]) is not None
                and self.twBehaviors.item(row, cfg.behavioursFields["coding map"]).text()
            ):
                QMessageBox.warning(self, cfg.programName, "Use the coding map to set/modify the areas")
            else:
                subjects_list = []
                for subject_row in range(self.twSubjects.rowCount()):
                    key = self.twSubjects.item(subject_row, 0).text() if self.twSubjects.item(subject_row, 0) else ""
                    subjectName = self.twSubjects.item(subject_row, 1).text().strip() if self.twSubjects.item(subject_row, 1) else ""
                    subjects_list.append((subjectName, key))

                addModifierWindow = add_modifier.addModifierDialog(self.twBehaviors.item(row, column).text(), subjects=subjects_list)
                addModifierWindow.setWindowTitle(f'Set modifiers for "{self.twBehaviors.item(row, 2).text()}" behavior')
                if addModifierWindow.exec_():
                    self.twBehaviors.item(row, column).setText(addModifierWindow.getModifiers())

    def behavior_type_doubleclicked(self, row):
        """
        select type for behavior
        """

        if self.twBehaviors.item(row, cfg.behavioursFields[cfg.TYPE]).text() in cfg.BEHAVIOR_TYPES:
            selected = cfg.BEHAVIOR_TYPES.index(self.twBehaviors.item(row, cfg.behavioursFields[cfg.TYPE]).text())
        else:
            selected = 0

        new_type, ok = QInputDialog.getItem(self, "Select a behavior type", "Types of behavior", cfg.BEHAVIOR_TYPES, selected, False)

        if ok and new_type:
            self.twBehaviors.item(row, cfg.behavioursFields["type"]).setText(new_type)

            self.behaviorTypeChanged(row)

    def color_doubleclicked(self, row: int) -> None:
        """
        select a color for behavior
        Selecting black delete the color
        """

        col_diag = QColorDialog()
        col_diag.setOptions(QColorDialog.ShowAlphaChannel | QColorDialog.DontUseNativeDialog)

        if self.twBehaviors.item(row, cfg.behavioursFields[cfg.COLOR]).text():
            current_color = QColor(self.twBehaviors.item(row, cfg.behavioursFields[cfg.COLOR]).text())
            if current_color.isValid():
                col_diag.setCurrentColor(current_color)

        if col_diag.exec_():
            color = col_diag.currentColor()
            if color.name() == "#000000":  # black -> delete color
                self.twBehaviors.item(row, cfg.behavioursFields[cfg.COLOR]).setText("")
                self.twBehaviors.item(row, cfg.behavioursFields[cfg.COLOR]).setBackground(QColor(230, 230, 230))
            elif color.isValid():
                self.twBehaviors.item(row, cfg.behavioursFields[cfg.COLOR]).setText(color.name())
                self.twBehaviors.item(row, cfg.behavioursFields[cfg.COLOR]).setBackground(color)

    def category_doubleclicked(self, row):
        """
        select category for behavior
        """

        categories = ["None"] + self.pj[cfg.BEHAVIORAL_CATEGORIES] if cfg.BEHAVIORAL_CATEGORIES in self.pj else ["None"]

        if self.twBehaviors.item(row, cfg.behavioursFields[cfg.BEHAVIOR_CATEGORY]).text() in categories:
            selected = categories.index(self.twBehaviors.item(row, cfg.behavioursFields[cfg.BEHAVIOR_CATEGORY]).text())
        else:
            selected = 0

        category, ok = QInputDialog.getItem(self, "Select a behavioral category", "Behavioral categories", categories, selected, False)

        if ok and category:
            if category == "None":
                category = ""
            self.twBehaviors.item(row, cfg.behavioursFields[cfg.BEHAVIOR_CATEGORY]).setText(category)

    def check_variable_default_value(self, txt, varType):
        """
        check if variable default value is compatible with variable type
        """
        # check for numeric type
        if varType == cfg.NUMERIC:
            try:
                if txt:
                    float(txt)
                return True
            except Exception:
                return False

        return True

    def variableTypeChanged(self, row):
        """
        variable type combobox changed
        """

        if self.twVariables.cellWidget(row, cfg.tw_indVarFields.index("type")).currentText() == cfg.SET_OF_VALUES:
            if self.twVariables.item(row, cfg.tw_indVarFields.index("possible values")).text() == "NA":
                self.twVariables.item(row, cfg.tw_indVarFields.index("possible values")).setText("Double-click to add values")
        else:
            # check if set of values defined
            if self.twVariables.item(row, cfg.tw_indVarFields.index("possible values")).text() not in [
                "NA",
                "Double-click to add values",
            ]:
                if dialog.MessageDialog(cfg.programName, "Erase the set of values?", [cfg.YES, cfg.CANCEL]) == cfg.CANCEL:
                    self.twVariables.cellWidget(row, cfg.tw_indVarFields.index("type")).setCurrentIndex(cfg.SET_OF_VALUES_idx)
                    return
                else:
                    self.twVariables.item(row, cfg.tw_indVarFields.index("possible values")).setText("NA")
            else:
                self.twVariables.item(row, cfg.tw_indVarFields.index("possible values")).setText("NA")

        # check compatibility between variable type and default value
        if not self.check_variable_default_value(
            self.twVariables.item(row, cfg.tw_indVarFields.index("default value")).text(),
            self.twVariables.cellWidget(row, cfg.tw_indVarFields.index("type")).currentIndex(),
        ):
            QMessageBox.warning(
                self,
                cfg.programName + " - Independent variables error",
                (
                    f"The default value ({self.twVariables.item(row, cfg.tw_indVarFields.index('default value')).text()}) "
                    f"of variable <b>{self.twVariables.item(row, cfg.tw_indVarFields.index('label')).text()}</b> "
                    "is not compatible with variable type"
                ),
            )

    def check_indep_var_config(self) -> tuple:
        """
        check if default type is compatible with var type
        """

        existing_var = []
        for r in range(self.twVariables.rowCount()):
            if self.twVariables.item(r, 0).text().strip().upper() in existing_var:
                return (
                    False,
                    f"Row: {r + 1} - The variable label <b>{self.twVariables.item(r, 0).text()}</b> is already in use.",
                )

            # check if same lables
            existing_var.append(self.twVariables.item(r, 0).text().strip().upper())

            # check default value
            if self.twVariables.item(r, 2).text() != cfg.TIMESTAMP and not self.check_variable_default_value(
                self.twVariables.item(r, 3).text(), self.twVariables.item(r, 2).text()
            ):
                return False, (
                    f"Row: {r + 1} - "
                    f"The default value ({self.twVariables.item(r, 3).text()}) is not compatible "
                    f"with the variable type ({self.twVariables.item(r, 2).text()})"
                )

            # check if default value in set of values
            if self.twVariables.item(r, 2).text() == cfg.SET_OF_VALUES and self.twVariables.item(r, 4).text() == "":
                return False, "No values were defined in set"

            if (
                self.twVariables.item(r, 2).text() == cfg.SET_OF_VALUES
                and self.twVariables.item(r, 4).text()
                and self.twVariables.item(r, 3).text()
                and self.twVariables.item(r, 3).text() not in self.twVariables.item(r, 4).text().split(",")
            ):
                return (
                    False,
                    f"The default value ({self.twVariables.item(r, 3).text()}) is not contained in set of values",
                )

        return True, "OK"

    def cbtype_changed(self):
        self.leSetValues.setVisible(self.cbType.currentText() == cfg.SET_OF_VALUES)
        self.label_5.setVisible(self.cbType.currentText() == cfg.SET_OF_VALUES)

        self.dte_default_date.setVisible(self.cbType.currentText() == cfg.TIMESTAMP)
        self.label_9.setVisible(self.cbType.currentText() == cfg.TIMESTAMP)
        self.lePredefined.setVisible(self.cbType.currentText() != cfg.TIMESTAMP)
        self.label_4.setVisible(self.cbType.currentText() != cfg.TIMESTAMP)

    def cbtype_activated(self):
        if self.cbType.currentText() == cfg.TIMESTAMP:
            self.twVariables.item(self.selected_twvariables_row, 3).setText(
                self.dte_default_date.dateTime().toString("yyyy-MM-ddTHH:mm:ss.zzz")
            )
            self.twVariables.item(self.selected_twvariables_row, 4).setText("")
        else:
            self.twVariables.item(self.selected_twvariables_row, 3).setText(self.lePredefined.text())
            self.twVariables.item(self.selected_twvariables_row, 4).setText("")

        # remove spaces after and before comma
        if self.cbType.currentText() == cfg.SET_OF_VALUES:
            self.twVariables.item(self.selected_twvariables_row, 4).setText(
                ",".join([x.strip() for x in self.leSetValues.text().split(",")])
            )

        self.twVariables.item(self.selected_twvariables_row, 2).setText(self.cbType.currentText())

        r, msg = self.check_indep_var_config()

        if not r:
            QMessageBox.warning(self, f"{cfg.programName} - Independent variables error", msg)

    def pbAddVariable_clicked(self):
        """
        add an independent variable
        """

        logging.debug("add an independent variable")
        self.twVariables.setRowCount(self.twVariables.rowCount() + 1)
        self.selected_twvariables_row = self.twVariables.rowCount() - 1

        for idx, field in enumerate(cfg.tw_indVarFields):
            if field == "type":
                item = QTableWidgetItem("numeric")
            else:
                item = QTableWidgetItem("")
            self.twVariables.setItem(self.twVariables.rowCount() - 1, idx, item)

        self.twVariables.setCurrentCell(self.twVariables.rowCount() - 1, 0)

        self.twVariables_cellClicked(self.twVariables.rowCount() - 1, 0)

    def pbRemoveVariable_clicked(self):
        """
        remove the selected independent variable
        """
        logging.debug("remove selected independent variable")

        if not self.twVariables.selectedIndexes():
            QMessageBox.warning(self, cfg.programName, "Select a variable to remove")
        else:
            if dialog.MessageDialog(cfg.programName, "Remove the selected variable?", [cfg.YES, cfg.CANCEL]) == cfg.YES:
                self.twVariables.removeRow(self.twVariables.selectedIndexes()[0].row())

        if self.twVariables.selectedIndexes():
            self.twVariables_cellClicked(self.twVariables.selectedIndexes()[0].row(), 0)
        else:
            self.twVariables_cellClicked(-1, 0)

    def pbImportVarFromProject_clicked(self):
        """
        import independent variables from another project
        """

        project_import_export.import_indep_variables_from_project(self)

    def exclusion_matrix(self):
        """
        activate exclusion matrix window
        """

        if not self.twBehaviors.rowCount():
            QMessageBox.critical(
                None,
                cfg.programName,
                "The ethogram is empty",
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )
            return

        for row in range(self.twBehaviors.rowCount()):
            if not self.twBehaviors.item(row, cfg.behavioursFields[cfg.BEHAVIOR_CODE]).text():
                QMessageBox.critical(
                    None,
                    cfg.programName,
                    f"A behavior code is empty at row {row + 1}",
                    QMessageBox.Ok | QMessageBox.Default,
                    QMessageBox.NoButton,
                )
                return

        ex = exclusion_matrix.ExclusionMatrix()

        state_behaviors, point_behaviors, allBehaviors, excl, new_excl = [], [], [], {}, {}

        # list of point events
        for r in range(self.twBehaviors.rowCount()):
            if self.twBehaviors.item(r, cfg.behavioursFields[cfg.BEHAVIOR_CODE]):
                if "Point" in self.twBehaviors.item(r, cfg.behavioursFields[cfg.TYPE]).text():
                    point_behaviors.append(self.twBehaviors.item(r, cfg.behavioursFields[cfg.BEHAVIOR_CODE]).text())

        # check if point are present and if user want to include them in exclusion matrix
        include_point_events = cfg.NO
        if point_behaviors:
            include_point_events = dialog.MessageDialog(
                cfg.programName,
                "Do you want to include the point events in the exclusion matrix?",
                [cfg.YES, cfg.NO],
            )

        for r in range(self.twBehaviors.rowCount()):
            if self.twBehaviors.item(r, cfg.behavioursFields[cfg.BEHAVIOR_CODE]):
                if include_point_events == cfg.YES or (
                    include_point_events == cfg.NO and "State" in self.twBehaviors.item(r, cfg.behavioursFields[cfg.TYPE]).text()
                ):
                    allBehaviors.append(self.twBehaviors.item(r, cfg.behavioursFields[cfg.BEHAVIOR_CODE]).text())

                excl[self.twBehaviors.item(r, cfg.behavioursFields[cfg.BEHAVIOR_CODE]).text()] = (
                    self.twBehaviors.item(r, cfg.behavioursFields["excluded"]).text().split(",")
                )
                new_excl[self.twBehaviors.item(r, cfg.behavioursFields[cfg.BEHAVIOR_CODE]).text()] = []

                if "State" in self.twBehaviors.item(r, cfg.behavioursFields[cfg.TYPE]).text():
                    state_behaviors.append(self.twBehaviors.item(r, cfg.behavioursFields[cfg.BEHAVIOR_CODE]).text())

        logging.debug(f"point behaviors: {point_behaviors}")
        logging.debug(f"state behaviors: {state_behaviors}")

        if not state_behaviors:
            QMessageBox.critical(
                None,
                cfg.programName,
                "No state events were defined in ethogram",
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )
            return

        logging.debug(f"exclusion matrix {excl}")

        # first row contain state events
        ex.twExclusions.setColumnCount(len(state_behaviors))
        ex.twExclusions.setHorizontalHeaderLabels(state_behaviors)
        ex.twExclusions.horizontalHeader().setSectionResizeMode(QHeaderView.Fixed)

        # first column contains all events: point + state
        ex.twExclusions.setRowCount(len(point_behaviors + state_behaviors))
        for idx, header in enumerate(point_behaviors + state_behaviors):
            item = QTableWidgetItem(header)
            if header in point_behaviors:
                item.setBackground(QColor(0, 200, 200))
            ex.twExclusions.setVerticalHeaderItem(idx, item)
        ex.twExclusions.verticalHeader().setSectionResizeMode(QHeaderView.Fixed)

        ex.allBehaviors = allBehaviors
        ex.stateBehaviors = state_behaviors
        ex.point_behaviors = point_behaviors

        ex.checkboxes = {}

        for c, c_name in enumerate(state_behaviors):
            flag_left_bottom = False
            for r, r_name in enumerate(point_behaviors + state_behaviors):
                if c_name == r_name:
                    flag_left_bottom = True

                if c_name != r_name:
                    ex.checkboxes[f"{r_name}|{c_name}"] = QCheckBox()
                    ex.checkboxes[f"{r_name}|{c_name}"].setStyleSheet("text-align: center; margin-left:50%; margin-right:50%;")

                    if flag_left_bottom:
                        # hide if cell in left-bottom part of table
                        ex.checkboxes[f"{r_name}|{c_name}"].setEnabled(False)

                    # connect function when a CB is clicked
                    ex.checkboxes[f"{r_name}|{c_name}"].clicked.connect(ex.cb_clicked)
                    if c_name in excl[r_name]:
                        ex.checkboxes[f"{r_name}|{c_name}"].setChecked(True)
                    ex.twExclusions.setCellWidget(r, c, ex.checkboxes[f"{r_name}|{c_name}"])

        ex.twExclusions.resizeColumnsToContents()
        # check corresponding checkbox
        ex.cb_clicked()

        if ex.exec_():
            for c, c_name in enumerate(state_behaviors):
                for r, r_name in enumerate(point_behaviors + state_behaviors):
                    if c_name != r_name:
                        if ex.twExclusions.cellWidget(r, c).isChecked():
                            if c_name not in new_excl[r_name]:
                                new_excl[r_name].append(c_name)

            logging.debug(f"new exclusion matrix {new_excl}")

            # update excluded field
            for r in range(self.twBehaviors.rowCount()):
                if include_point_events == cfg.YES or (include_point_events == cfg.NO and "State" in self.twBehaviors.item(r, 0).text()):
                    for e in excl:
                        if e == self.twBehaviors.item(r, cfg.behavioursFields[cfg.BEHAVIOR_CODE]).text():
                            item = QTableWidgetItem(",".join(new_excl[e]))
                            item.setFlags(Qt.ItemIsEnabled)
                            item.setBackground(QColor(230, 230, 230))
                            self.twBehaviors.setItem(r, cfg.behavioursFields["excluded"], item)

    def remove_all_behaviors(self):
        if not self.twBehaviors.rowCount():
            QMessageBox.critical(
                None,
                cfg.programName,
                "The ethogram is empty",
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )
            return

        if dialog.MessageDialog(cfg.programName, "Remove all behaviors?", [cfg.YES, cfg.CANCEL]) != cfg.YES:
            return

        # delete ethogram rows without behavior code
        for r in range(self.twBehaviors.rowCount() - 1, -1, -1):
            if not self.twBehaviors.item(r, cfg.PROJECT_BEHAVIORS_CODE_FIELD_IDX).text():
                self.twBehaviors.removeRow(r)

        # extract all codes to delete
        codesToDelete = []
        row_mem = {}
        for r in range(self.twBehaviors.rowCount() - 1, -1, -1):
            if self.twBehaviors.item(r, cfg.PROJECT_BEHAVIORS_CODE_FIELD_IDX).text():
                codesToDelete.append(self.twBehaviors.item(r, cfg.PROJECT_BEHAVIORS_CODE_FIELD_IDX).text())
                row_mem[self.twBehaviors.item(r, cfg.PROJECT_BEHAVIORS_CODE_FIELD_IDX).text()] = r

        # extract all codes used in observations
        codesInObs = []
        for obs in self.pj[cfg.OBSERVATIONS]:
            events = self.pj[cfg.OBSERVATIONS][obs][cfg.EVENTS]
            for event in events:
                codesInObs.append(event[cfg.EVENT_BEHAVIOR_FIELD_IDX])

        for codeToDelete in codesToDelete:
            # if code to delete used in obs ask confirmation
            if codeToDelete in codesInObs:
                response = dialog.MessageDialog(
                    cfg.programName,
                    f"The code <b>{codeToDelete}</b> is used in observations!",
                    ["Remove", cfg.CANCEL],
                )
                if response == "Remove":
                    self.twBehaviors.removeRow(row_mem[codeToDelete])
            else:  # remove without asking
                self.twBehaviors.removeRow(row_mem[codeToDelete])

    def twBehaviors_cellChanged(self, row, column):
        """
        check ethogram
        """

        keys, codes = [], []
        self.lbObservationsState.setText("")

        for r in range(self.twBehaviors.rowCount()):
            # check key
            if self.twBehaviors.item(r, cfg.PROJECT_BEHAVIORS_KEY_FIELD_IDX):
                key = self.twBehaviors.item(r, cfg.PROJECT_BEHAVIORS_KEY_FIELD_IDX).text()
                # check key length
                if key.upper() not in list(cfg.function_keys.values()) and len(key) > 1:
                    self.lbObservationsState.setText('<font color="red">Key length &gt; 1</font>')
                    return

                keys.append(key)

            # check code
            if self.twBehaviors.item(r, cfg.PROJECT_BEHAVIORS_CODE_FIELD_IDX):
                if self.twBehaviors.item(r, cfg.PROJECT_BEHAVIORS_CODE_FIELD_IDX).text() in codes:
                    self.lbObservationsState.setText(f'<font color="red">Code duplicated at line {r + 1} </font>')
                else:
                    if self.twBehaviors.item(r, cfg.PROJECT_BEHAVIORS_CODE_FIELD_IDX).text():
                        codes.append(self.twBehaviors.item(r, cfg.PROJECT_BEHAVIORS_CODE_FIELD_IDX).text())

    def clone_behavior(self):
        """
        clone the selected behavior
        """

        if not self.twBehaviors.rowCount():
            QMessageBox.critical(
                None,
                cfg.programName,
                "The ethogram is empty",
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )
            return

        if not self.twBehaviors.selectedIndexes():
            QMessageBox.about(self, cfg.programName, "First select a behavior")
        else:
            self.twBehaviors.setRowCount(self.twBehaviors.rowCount() + 1)

            row = self.twBehaviors.selectedIndexes()[0].row()
            for field in cfg.behavioursFields:
                item = QTableWidgetItem(self.twBehaviors.item(row, cfg.behavioursFields[field]))
                self.twBehaviors.setItem(self.twBehaviors.rowCount() - 1, cfg.behavioursFields[field], item)
                if field in (cfg.TYPE, "category", "excluded", "coding map", "modifiers"):
                    item.setFlags(Qt.ItemIsEnabled)
                    item.setBackground(QColor(230, 230, 230))
                if field == cfg.COLOR:
                    item.setFlags(Qt.ItemIsEnabled)
                    if QColor(self.twBehaviors.item(row, cfg.behavioursFields[field]).text()).isValid():
                        item.setBackground(QColor(self.twBehaviors.item(row, cfg.behavioursFields[field]).text()))
                    else:
                        item.setBackground(QColor(230, 230, 230))
        self.twBehaviors.scrollToBottom()

    def remove_behavior(self):
        """
        remove behavior
        """

        if not self.twBehaviors.rowCount():
            QMessageBox.critical(
                None,
                cfg.programName,
                "The ethogram is empty",
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )
            return

        if not self.twBehaviors.selectedIndexes():
            QMessageBox.warning(self, cfg.programName, "Select a behaviour to be removed")
            return

        if dialog.MessageDialog(cfg.programName, "Remove the selected behavior?", [cfg.YES, cfg.CANCEL]) == cfg.YES:
            # check if behavior already used in observations
            codeToDelete = self.twBehaviors.item(self.twBehaviors.selectedIndexes()[0].row(), 2).text()
            for obs_id in self.pj[cfg.OBSERVATIONS]:
                if codeToDelete in [event[cfg.EVENT_BEHAVIOR_FIELD_IDX] for event in self.pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS]]:
                    if (
                        dialog.MessageDialog(cfg.programName, "The code to remove is used in observations!", [cfg.REMOVE, cfg.CANCEL])
                        == cfg.CANCEL
                    ):
                        return
                    break

            self.twBehaviors.removeRow(self.twBehaviors.selectedIndexes()[0].row())
            self.twBehaviors_cellChanged(0, 0)

    def add_behavior(self):
        """
        add new behavior to ethogram
        """

        # Add behavior to table
        self.twBehaviors.setRowCount(self.twBehaviors.rowCount() + 1)
        for field_type in cfg.behavioursFields:
            item = QTableWidgetItem()
            if field_type == cfg.TYPE:
                item.setText("Point event")
            # no manual editing, gray back ground
            if field_type in (cfg.TYPE, cfg.COLOR, "category", cfg.MODIFIERS, "modifiers", "excluded", "coding map"):
                item.setFlags(Qt.ItemIsEnabled)
                item.setBackground(QColor(230, 230, 230))
            self.twBehaviors.setItem(self.twBehaviors.rowCount() - 1, cfg.behavioursFields[field_type], item)
        self.twBehaviors.scrollToBottom()

    def behaviorTypeChanged(self, row):
        """
        event type combobox changed
        """

        if "with coding map" in self.twBehaviors.item(row, cfg.behavioursFields[cfg.TYPE]).text():
            # let user select a coding maop
            fn = QFileDialog().getOpenFileName(
                self,
                "Select a coding map for " f"{self.twBehaviors.item(row, cfg.behavioursFields['code']).text()} behavior",
                "",
                "BORIS map files (*.boris_map);;All files (*)",
            )
            fileName = fn[0] if type(fn) is tuple else fn

            if fileName:
                try:
                    new_map = json.loads(open(fileName, "r").read())
                except Exception:
                    QMessageBox.critical(self, cfg.programName, "Error reding the file")
                    return
                self.pj[cfg.CODING_MAP][new_map["name"]] = new_map

                # add modifiers from coding map areas
                modifstr = str(
                    {
                        "0": {
                            "name": new_map["name"],
                            "type": cfg.MULTI_SELECTION,
                            "values": list(sorted(new_map["areas"].keys())),
                        }
                    }
                )

                self.twBehaviors.item(row, cfg.behavioursFields["modifiers"]).setText(modifstr)
                self.twBehaviors.item(row, cfg.behavioursFields["coding map"]).setText(new_map["name"])

            else:
                # if coding map already exists do not reset the behavior type if no filename selected
                if not self.twBehaviors.item(row, cfg.behavioursFields["coding map"]).text():
                    QMessageBox.critical(self, cfg.programName, 'No coding map was selected.\nEvent type will be reset to "Point event" ')
                    self.twBehaviors.item(row, cfg.behavioursFields["type"]).setText("Point event")
        else:
            self.twBehaviors.item(row, cfg.behavioursFields["coding map"]).setText("")

    def add_subject(self):
        """
        add a subject
        """

        self.twSubjects.setRowCount(self.twSubjects.rowCount() + 1)
        for col in range(len(cfg.subjectsFields)):
            item = QTableWidgetItem("")
            self.twSubjects.setItem(self.twSubjects.rowCount() - 1, col, item)
        self.twSubjects.scrollToBottom()

    def remove_subject(self):
        """
        remove selected subject from subjects list
        control before if subject used in observations
        """

        if not self.twSubjects.selectedIndexes():
            QMessageBox.warning(self, cfg.programName, "Select a subject to remove")
        else:
            if dialog.MessageDialog(cfg.programName, "Remove the selected subject?", [cfg.YES, cfg.CANCEL]) == cfg.YES:
                flagDel = False
                if self.twSubjects.item(self.twSubjects.selectedIndexes()[0].row(), 1):
                    subjectToDelete = self.twSubjects.item(self.twSubjects.selectedIndexes()[0].row(), 1).text()  # 1: subject name

                    subjectsInObs = []
                    for obs in self.pj[cfg.OBSERVATIONS]:
                        events = self.pj[cfg.OBSERVATIONS][obs][cfg.EVENTS]
                        for event in events:
                            subjectsInObs.append(event[cfg.EVENT_SUBJECT_FIELD_IDX])
                    if subjectToDelete in subjectsInObs:
                        if (
                            dialog.MessageDialog(
                                cfg.programName,
                                "The subject to remove is used in observations!",
                                [cfg.REMOVE, cfg.CANCEL],
                            )
                            == cfg.REMOVE
                        ):
                            flagDel = True
                    else:
                        # code not used
                        flagDel = True
                else:
                    flagDel = True

                if flagDel:
                    self.twSubjects.removeRow(self.twSubjects.selectedIndexes()[0].row())

                self.twSubjects_cellChanged(0, 0)

    def remove_all_subjects(self):
        """
        remove all subjects.
        Verify if they are used in observations
        """

        if not self.twSubjects.rowCount():
            return

        if dialog.MessageDialog(cfg.programName, "Remove all subjects?", [cfg.YES, cfg.CANCEL]) != cfg.YES:
            return

        # delete ethogram rows without behavior code
        for r in range(self.twSubjects.rowCount() - 1, -1, -1):
            if not self.twSubjects.item(r, 1).text():  # no name
                self.twSubjects.removeRow(r)

        # extract all subjects names to delete
        namesToDelete: list = []
        row_mem: dict = {}
        for r in range(self.twSubjects.rowCount() - 1, -1, -1):
            if self.twSubjects.item(r, 1).text():
                namesToDelete.append(self.twSubjects.item(r, 1).text())
                row_mem[self.twSubjects.item(r, 1).text()] = r

        # extract all subjects name used in observations
        namesInObs: list = []
        for obs in self.pj[cfg.OBSERVATIONS]:
            events = self.pj[cfg.OBSERVATIONS][obs][cfg.EVENTS]
            for event in events:
                namesInObs.append(event[cfg.EVENT_SUBJECT_FIELD_IDX])

        flag_force: bool = False
        for nameToDelete in namesToDelete:
            # if name to delete used in obs ask confirmation
            if nameToDelete in namesInObs and not flag_force:
                response = dialog.MessageDialog(
                    cfg.programName,
                    f"The subject <b>{nameToDelete}</b> is used in observations!",
                    ["Force removing of all subjects", cfg.REMOVE, cfg.CANCEL],
                )
                if response == "Force removing of all subjects":
                    flag_force = True
                    self.twSubjects.removeRow(row_mem[nameToDelete])

                if response == cfg.REMOVE:
                    self.twSubjects.removeRow(row_mem[nameToDelete])
            else:  # remove without asking
                self.twSubjects.removeRow(row_mem[nameToDelete])

        self.twSubjects_cellChanged(0, 0)

    def twSubjects_cellChanged(self, row: int, column: int) -> None:
        """
        check if subject not unique
        """

        subjects: list = []
        """keys: list = []"""
        self.lbSubjectsState.setText("")

        for r in range(self.twSubjects.rowCount()):
            # check key
            if self.twSubjects.item(r, 0):
                # check key length
                if (
                    self.twSubjects.item(r, 0).text().upper() not in list(cfg.function_keys.values())
                    and len(self.twSubjects.item(r, 0).text()) > 1
                ):
                    self.lbSubjectsState.setText(
                        (
                            f'<font color="red">Error on key {self.twSubjects.item(r, 0).text()} for subject!</font>'
                            "The key is too long (keys must be of one character"
                            " except for function keys _F1, F2..._)"
                        )
                    )
                    return

                # control of duplicated key removed 2024-01-29
                """
                if self.twSubjects.item(r, 0).text() in keys:
                    self.lbSubjectsState.setText(f'<font color="red">Key duplicated at row # {r + 1}</font>')
                else:
                    if self.twSubjects.item(r, 0).text():
                        keys.append(self.twSubjects.item(r, 0).text())
                """

            # check subject
            if self.twSubjects.item(r, 1):
                if self.twSubjects.item(r, 1).text() in subjects:
                    self.lbSubjectsState.setText(f'<font color="red">Subject duplicated at row # {r + 1}</font>')
                else:
                    if self.twSubjects.item(r, 1).text():
                        subjects.append(self.twSubjects.item(r, 1).text())

    def twVariables_cellClicked(self, row, column):
        """
        check if variable default values are compatible with variable type
        """

        self.selected_twvariables_row = row
        logging.debug(f"selected row: {self.selected_twvariables_row}")

        if self.selected_twvariables_row == -1:
            for widget in (
                self.leLabel,
                self.leDescription,
                self.cbType,
                self.lePredefined,
                self.dte_default_date,
                self.leSetValues,
            ):
                widget.setEnabled(False)
                self.leLabel.setText("")
                self.leDescription.setText("")
                self.lePredefined.setText("")
                self.leSetValues.setText("")

                self.cbType.clear()
            return

        # enable widget for indep var setting
        for widget in (
            self.leLabel,
            self.leDescription,
            self.cbType,
            self.lePredefined,
            self.dte_default_date,
            self.leSetValues,
        ):
            widget.setEnabled(True)

        self.leLabel.setText(self.twVariables.item(row, 0).text())
        self.leDescription.setText(self.twVariables.item(row, 1).text())
        self.lePredefined.setText(self.twVariables.item(row, 3).text())
        self.leSetValues.setText(self.twVariables.item(row, 4).text())
        if self.twVariables.item(row, 2).text() == cfg.TIMESTAMP:
            if len(self.twVariables.item(row, 3).text()) == len("yyyy-MM-ddTHH:mm:ss.zzz"):
                datetime_format = "yyyy-MM-ddThh:mm:ss.zzz"
            if len(self.twVariables.item(row, 3).text()) == len("yyyy-MM-ddTHH:mm:ss"):
                datetime_format = "yyyy-MM-ddThh:mm:ss"

            self.dte_default_date.setDateTime(QDateTime.fromString(self.twVariables.item(row, 3).text(), datetime_format))

        self.cbType.clear()
        self.cbType.addItems(cfg.AVAILABLE_INDEP_VAR_TYPES)
        self.cbType.setCurrentIndex(cfg.NUMERIC_idx)

        self.cbType.setCurrentIndex(cfg.AVAILABLE_INDEP_VAR_TYPES.index(self.twVariables.item(row, 2).text()))

    def pbCancel_clicked(self):
        if self.flag_modified:
            if dialog.MessageDialog("BORIS", "The converters were modified. Are you sure to cancel?", [cfg.CANCEL, cfg.OK]) == cfg.OK:
                self.reject()
        else:
            self.reject()

    def check_ethogram(self) -> dict:
        """
        check ethogram for various parameter
        returns ethogram dict or {cfg.CANCEL: True"} in case of error

        """
        # store behaviors
        missing_data: list = []
        checked_ethogram: dict = {}

        # Ethogram
        # coding maps in ethogram

        # check for leading/trailing space in behaviors and modifiers
        code_with_leading_trailing_spaces, modifiers_with_leading_trailing_spaces = [], []
        for r in range(self.twBehaviors.rowCount()):
            if (
                self.twBehaviors.item(r, cfg.behavioursFields[cfg.BEHAVIOR_CODE]).text()
                != self.twBehaviors.item(r, cfg.behavioursFields[cfg.BEHAVIOR_CODE]).text().strip()
            ):
                code_with_leading_trailing_spaces.append(self.twBehaviors.item(r, cfg.behavioursFields[cfg.BEHAVIOR_CODE]).text())

            if self.twBehaviors.item(r, cfg.behavioursFields["modifiers"]).text():
                try:
                    modifiers_dict = eval(self.twBehaviors.item(r, cfg.behavioursFields["modifiers"]).text())
                    for k in modifiers_dict:
                        for value in modifiers_dict[k]["values"]:
                            modif_code = value.split(" (")[0]
                            if modif_code.strip() != modif_code:
                                modifiers_with_leading_trailing_spaces.append(modif_code)
                except Exception:
                    logging.critical("error checking leading/trailing spaces in modifiers")

        remove_leading_trailing_spaces = cfg.NO
        if code_with_leading_trailing_spaces:
            remove_leading_trailing_spaces = dialog.MessageDialog(
                cfg.programName,
                (
                    "<b>Warning!</b> Some leading and/or trailing spaces are present"
                    " in the following behaviors code(s):<br>"
                    "<b>"
                    f"{'<br>'.join([util.replace_leading_trailing_chars(x, ' ', '&#9608;') for x in code_with_leading_trailing_spaces])}"
                    "</b><br><br>"
                    "Do you want to remove the leading and trailing spaces (visualized as black boxes) from behaviors?<br><br>"
                    """<font color="red"><b>Be careful with this option"""
                    """ if you have already done observations!</b></font>"""
                ),
                [cfg.YES, cfg.NO, cfg.CANCEL],
            )
        if remove_leading_trailing_spaces == cfg.CANCEL:
            return {cfg.CANCEL: True}

        remove_leading_trailing_spaces_in_modifiers = cfg.NO
        if modifiers_with_leading_trailing_spaces:
            remove_leading_trailing_spaces_in_modifiers = dialog.MessageDialog(
                cfg.programName,
                (
                    "<b>Warning!</b> Some leading and/or trailing spaces are present"
                    " in the following modifier(s):<br><b>"
                    f"{'<br>'.join([util.replace_leading_trailing_chars(x, ' ', '&#9608;') for x in set(modifiers_with_leading_trailing_spaces)])}"
                    "</b><br><br>Do you want to remove the leading and trailing spaces (visualized as black boxes) from modifiers?<br><br>"
                    """<font color="red"><b>Be careful with this option"""
                    """ if you have already done observations!</b></font>"""
                ),
                (cfg.YES, cfg.NO, cfg.CANCEL),
            )
        if remove_leading_trailing_spaces_in_modifiers == cfg.CANCEL:
            return {cfg.CANCEL: True}

        codingMapsList = []
        for r in range(self.twBehaviors.rowCount()):
            row = {}
            for field in cfg.behavioursFields:
                if self.twBehaviors.item(r, cfg.behavioursFields[field]):
                    # check for | char in code
                    if field == cfg.BEHAVIOR_CODE and "|" in self.twBehaviors.item(r, cfg.behavioursFields[field]).text():
                        QMessageBox.warning(
                            self,
                            cfg.programName,
                            (
                                "The pipe (|) character is not allowed in code "
                                f"<b>{self.twBehaviors.item(r, cfg.behavioursFields[field]).text()}</b> !"
                            ),
                        )
                        return {cfg.CANCEL: True}

                    if remove_leading_trailing_spaces == cfg.YES:
                        row[field] = self.twBehaviors.item(r, cfg.behavioursFields[field]).text().strip()
                    else:
                        row[field] = self.twBehaviors.item(r, cfg.behavioursFields[field]).text()

                    if field == "modifiers" and row["modifiers"]:
                        if remove_leading_trailing_spaces_in_modifiers == cfg.YES:
                            try:
                                modifiers_dict = eval(row["modifiers"])
                                for k in modifiers_dict:
                                    for idx, value in enumerate(modifiers_dict[k]["values"]):
                                        modif_code = value.split(" (")[0]

                                        modifiers_dict[k]["values"][idx] = modifiers_dict[k]["values"][idx].replace(
                                            modif_code, modif_code.strip()
                                        )

                                row["modifiers"] = dict(modifiers_dict)
                            except Exception:
                                logging.critical("Error removing leading/trailing spaces in modifiers")

                                QMessageBox.critical(self, cfg.programName, "Error removing leading/trailing spaces in modifiers")

                        else:
                            row["modifiers"] = eval(row["modifiers"])
                else:
                    row[field] = ""

            if (row["type"]) and (row[cfg.BEHAVIOR_CODE]):
                checked_ethogram[str(len(checked_ethogram))] = row
            else:
                missing_data.append(str(r + 1))

            if self.twBehaviors.item(r, cfg.behavioursFields["coding map"]).text():
                codingMapsList.append(self.twBehaviors.item(r, cfg.behavioursFields["coding map"]).text())

        # remove coding map from project if not in ethogram
        cmToDelete = []
        for cm in self.pj[cfg.CODING_MAP]:
            if cm not in codingMapsList:
                cmToDelete.append(cm)

        for cm in cmToDelete:
            del self.pj[cfg.CODING_MAP][cm]

        if missing_data:
            QMessageBox.warning(self, cfg.programName, f"Missing data in ethogram at row {','.join(missing_data)} !")
            return {cfg.CANCEL: True}

        # check if behavior belong to category that is not in categories list
        behavior_category = []
        for idx in checked_ethogram:
            if cfg.BEHAVIOR_CATEGORY in checked_ethogram[idx]:
                if checked_ethogram[idx][cfg.BEHAVIOR_CATEGORY]:
                    if checked_ethogram[idx][cfg.BEHAVIOR_CATEGORY] not in self.pj[cfg.BEHAVIORAL_CATEGORIES]:
                        behavior_category.append((checked_ethogram[idx][cfg.BEHAVIOR_CODE], checked_ethogram[idx][cfg.BEHAVIOR_CATEGORY]))
        if behavior_category:
            response = dialog.MessageDialog(
                f"{cfg.programName} - Behavioral categories",
                (
                    "The behavioral categorie(s) "
                    f"{', '.join(set(['<b>' + x[1] + '</b>' + ' (used with <b>' + x[0] + '</b>)' for x in behavior_category]))} "
                    "are no more defined in behavioral categories list"
                ),
                ["Add behavioral category/ies", "Ignore", cfg.CANCEL],
            )
            if response == "Add behavioral category/ies":
                [self.pj[cfg.BEHAVIORAL_CATEGORIES].append(x1) for x1 in set(x[1] for x in behavior_category)]
            if response == cfg.CANCEL:
                return {cfg.CANCEL: True}

        # delete coding maps loaded in pj and not cited in ethogram
        return checked_ethogram

    def pbOK_clicked(self):
        """
        verify project configuration
        """

        if self.lbObservationsState.text():
            QMessageBox.warning(self, cfg.programName, self.lbObservationsState.text())
            return

        if self.lbSubjectsState.text():
            QMessageBox.warning(self, cfg.programName, self.lbSubjectsState.text())
            return

        self.pj[cfg.PROJECT_NAME] = self.leProjectName.text().strip()
        self.pj[cfg.PROJECT_DATE] = self.dteDate.dateTime().toString(Qt.ISODate)
        self.pj[cfg.PROJECT_DESCRIPTION] = self.teDescription.toPlainText()

        # time format
        if self.rbSeconds.isChecked():
            self.pj[cfg.TIME_FORMAT] = cfg.S
        if self.rbHMS.isChecked():
            self.pj[cfg.TIME_FORMAT] = cfg.HHMMSS

        # store subjects
        self.subjects_conf: dict = {}

        # check for leading/trailing spaces in subjects names
        subjects_name_with_leading_trailing_spaces = ""
        for row in range(self.twSubjects.rowCount()):
            if self.twSubjects.item(row, 1):
                if self.twSubjects.item(row, 1).text() != self.twSubjects.item(row, 1).text().strip():
                    subjects_name_with_leading_trailing_spaces += f'"{self.twSubjects.item(row, 1).text()}" '

        remove_leading_trailing_spaces = cfg.NO
        if subjects_name_with_leading_trailing_spaces:
            remove_leading_trailing_spaces = dialog.MessageDialog(
                cfg.programName,
                (
                    "Attention! Some leading and/or trailing spaces are present in the following <b>subject name(s)</b>:<br>"
                    f"<b>{subjects_name_with_leading_trailing_spaces}</b><br><br>"
                    "Do you want to remove the leading and trailing spaces?<br><br>"
                    '<font color="red"><b>Be careful with this option'
                    " if you have already done observations!</b></font>"
                ),
                [cfg.YES, cfg.NO],
            )

        # check subjects
        for row in range(self.twSubjects.rowCount()):
            # check key
            if self.twSubjects.item(row, 0):
                key = self.twSubjects.item(row, 0).text()
            else:
                key = ""

            # check subject name
            if self.twSubjects.item(row, 1):
                if remove_leading_trailing_spaces == cfg.YES:
                    subjectName = self.twSubjects.item(row, 1).text().strip()
                else:
                    subjectName = self.twSubjects.item(row, 1).text()

                # check if subject name is empty
                if subjectName == "":
                    QMessageBox.warning(self, cfg.programName, f"The subject name can not be empty (check row #{row + 1}).")
                    return

                if "|" in subjectName:
                    QMessageBox.warning(
                        self,
                        cfg.programName,
                        f"The pipe (|) character is not allowed in subject name <b>{subjectName}</b>",
                    )
                    return
            else:
                QMessageBox.warning(self, cfg.programName, f"Missing subject name in subjects configuration at row #{row + 1}")
                return

            # description
            subjectDescription = ""
            if self.twSubjects.item(row, 2):
                subjectDescription = self.twSubjects.item(row, 2).text().strip()

            self.subjects_conf[str(len(self.subjects_conf))] = {
                "key": key,
                "name": subjectName,
                "description": subjectDescription,
            }

        self.pj[cfg.SUBJECTS] = dict(self.subjects_conf)

        # check ethogram
        r = dict(self.check_ethogram())
        if cfg.CANCEL in r:
            return
        self.pj[cfg.ETHOGRAM] = dict(r)

        # independent variables
        r, msg = self.check_indep_var_config()
        if not r:
            QMessageBox.warning(self, cfg.programName + " - Independent variables error", msg)
            return

        self.indVar = {}
        for r in range(self.twVariables.rowCount()):
            row = {}
            for idx, field in enumerate(cfg.tw_indVarFields):
                if self.twVariables.item(r, idx):
                    # check if label is empty
                    if field == "label" and self.twVariables.item(r, idx).text() == "":
                        QMessageBox.warning(
                            self,
                            cfg.programName,
                            f"The label of an indipendent variable can not be empty (check row #{r + 1}).",
                        )
                        return

                    row[field] = self.twVariables.item(r, idx).text().strip()
                else:
                    row[field] = ""

            self.indVar[str(len(self.indVar))] = row

        self.pj[cfg.INDEPENDENT_VARIABLES] = dict(self.indVar)

        # converters
        converters = {}
        for row in range(self.tw_converters.rowCount()):
            converters[self.tw_converters.item(row, 0).text()] = {
                "name": self.tw_converters.item(row, 0).text(),
                "description": self.tw_converters.item(row, 1).text(),
                "code": self.tw_converters.item(row, 2).text().replace("@", "\n"),
            }
        self.pj[cfg.CONVERTERS] = dict(converters)

        self.accept()

    def load_converters_in_table(self):
        """
        load converters in table
        """
        self.tw_converters.setRowCount(0)

        for converter in sorted(self.converters.keys()):
            self.tw_converters.setRowCount(self.tw_converters.rowCount() + 1)
            self.tw_converters.setItem(self.tw_converters.rowCount() - 1, 0, QTableWidgetItem(converter))  # id / name
            self.tw_converters.setItem(self.tw_converters.rowCount() - 1, 1, QTableWidgetItem(self.converters[converter]["description"]))
            self.tw_converters.setItem(
                self.tw_converters.rowCount() - 1,
                2,
                QTableWidgetItem(self.converters[converter]["code"].replace("\n", "@")),
            )

        [self.tw_converters.resizeColumnToContents(idx) for idx in [0, 1]]
